# Copyright(C) 2008 Brazil
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# TODO: add nogil

import atexit

cdef extern from "stdlib.h":
  ctypedef unsigned size_t # avoid...
  void *malloc(size_t size)
  void free(void *ptr)

cdef extern from "Python.h":
  int PyString_Size(object string)
  char *PyString_AsString(object obj)
  object PyString_FromStringAndSize(char *v, Py_ssize_t len)
  object PyString_FromString(char *v)
  bint PySequence_Check(object o)
  Py_ssize_t PySequence_Size(object o)

cdef extern from "senna/senna.h":
  ctypedef unsigned sen_id
  ctypedef enum sen_rc:
    sen_success = 0
    sen_memory_exhausted
    sen_invalid_format
    sen_file_operation_error
    sen_invalid_argument
    sen_other_error
    sen_external_error
    sen_internal_error
    sen_abnormal_error
    sen_end_of_data
  ctypedef enum sen_encoding:
    sen_enc_default = 0
    sen_enc_none
    sen_enc_euc_jp
    sen_enc_utf8
    sen_enc_sjis
    sen_enc_latin1
    sen_enc_koi8r
  ctypedef enum sen_sel_operator:
    sen_sel_or = 0
    sen_sel_and
    sen_sel_but
    sen_sel_adjust

  ctypedef struct sen_ctx:
    pass
  ctypedef struct sen_obj:
    pass
  ctypedef unsigned short int sen_obj_flags
  ctypedef struct sen_proc_ctx:
    pass

  sen_rc sen_init()
  sen_rc sen_fin()
  sen_rc sen_ctx_init(sen_ctx *ctx, int flags, sen_encoding encoding)
  sen_rc sen_ctx_fin(sen_ctx *ctx)
  sen_obj *sen_space_create(sen_ctx *ctx, char *path,
                            sen_encoding encoding)
  sen_obj *sen_space_open(sen_ctx *ctx, char *path)
  sen_obj *sen_space_lookup(sen_ctx *ctx, sen_obj *space,
                            char *name, unsigned name_size)
  extern sen_obj *sen_type_any
  sen_obj *sen_type_create(sen_ctx *ctx, sen_obj *space,
                           char *name, unsigned name_size,
                           sen_obj_flags flags, unsigned int size)

  ctypedef union sen_proc_data:
    int int_value
    sen_id id
    void *ptr
  ctypedef sen_rc sen_proc_func(sen_ctx *ctx, sen_proc_ctx *pctx,
                                int argc, sen_proc_data *argv)
  ctypedef sen_rc sen_proc_init_func(sen_ctx *ctx, sen_obj *space,
                                     char *path)
  ctypedef enum sen_proc_type:
    SEN_PROC_HOOK
    SEN_PROC_RECALC
    SEN_PROC_SCORE
    SEN_PROC_COMPARE
    SEN_PROC_GROUP

  cdef sen_obj *sen_proc_create(sen_ctx *ctx, sen_obj *space,
                                char *name, unsigned name_size,
                                char *path, sen_proc_type type,
                                sen_proc_func *init,
                                sen_proc_func *next,
                                sen_proc_func *fin)

  cdef sen_obj *sen_table_create(sen_ctx *ctx, sen_obj *space,
                                 char *name, unsigned name_size,
                                 char *path, sen_obj_flags flags,
                                 sen_obj *key_type, unsigned value_size,
                                 sen_encoding encoding)

  cdef sen_obj *sen_table_open(sen_ctx *ctx, sen_obj *space,
                               char *name, unsigned name_size, char *path)

  ctypedef unsigned char sen_search_flags
  cdef sen_id sen_table_lookup(sen_ctx *ctx, sen_obj *table,
                               void *key, unsigned key_size,
                               sen_search_flags *flags)
  cdef sen_id sen_table_add(sen_ctx *ctx, sen_obj *table)
  cdef int sen_table_get_key(sen_ctx *ctx, sen_obj *table,
                             sen_id id, void *keybuf, int buf_size)
  cdef sen_rc sen_table_delete(sen_ctx *ctx, sen_obj *table,
                               void *key, unsigned key_size)
  cdef sen_rc sen_table_delete_by_id(sen_ctx *ctx, sen_obj *table,
                                     sen_id id)
  cdef sen_rc sen_table_truncate(sen_ctx *ctx, sen_obj *table)

  ctypedef sen_obj sen_table_cursor

  ctypedef enum sen_update_type:
    SEN_UPD_SET = 0
    SEN_UPD_ADD
    SEN_UPD_CAS
    SEN_UPD_APPEND
    SEN_UPD_PREPEND

  cdef sen_table_cursor *sen_table_cursor_open(sen_ctx *ctx, sen_obj *table,
                                               void *min, unsigned min_size,
                                               void *max, unsigned max_size,
                                               int flags)

  cdef sen_rc sen_table_cursor_close(sen_ctx *ctx, sen_table_cursor *tc)
  cdef sen_id sen_table_cursor_next(sen_ctx *ctx, sen_table_cursor *tc)
  cdef int sen_table_cursor_get_key(sen_ctx *ctx, sen_table_cursor *tc,
                                    void **key)
  cdef int sen_table_cursor_get_value(sen_ctx *ctx, sen_table_cursor *tc,
                                      void **value)
  cdef sen_rc sen_table_cursor_set_value(sen_ctx *ctx, sen_table_cursor *tc,
                                         void *value, sen_update_type type)
  cdef sen_rc sen_table_cursor_delete(sen_ctx *ctx, sen_table_cursor *tc)

  ctypedef unsigned char sen_table_sort_flags

  ctypedef struct sen_table_sort_optarg:
    pass

  ctypedef struct sen_table_sort_optarg:
    sen_table_sort_flags flags
    int (*compar)(sen_obj *table1, void *target1, unsigned int target1_size,
                  sen_obj *table2, void *target2, unsigned int target2_size,
                  void *compare_arg)
    void *compar_arg
    sen_obj *proc
    int offset

  cdef int sen_table_sort(sen_ctx *ctx, sen_obj *table, int limit,
                          sen_obj *result, sen_table_sort_optarg *optarg)

  ctypedef struct sen_table_group_arg:
    sen_obj *result
    int flags
    sen_id slot
    int (*func)(sen_obj *, sen_id, void *)
    void *func_arg
    int limit
    sen_table_sort_optarg sortarg

  cdef sen_rc sen_table_group(sen_ctx *ctx, sen_obj *table,
                              int n_args, sen_table_group_arg *args)
  cdef sen_rc sen_table_setoperation(sen_ctx *ctx,
                                     sen_obj *table1, sen_obj *table2,
                                     sen_obj *res, sen_sel_operator op)
  cdef sen_rc sen_table_difference(sen_ctx *ctx,
                                   sen_obj *table1, sen_obj *table2,
                                   sen_obj *res1, sen_obj *res2)
  cdef sen_obj *sen_table_column(sen_ctx *ctx, sen_obj *table,
                                 char *name, unsigned name_size)
  cdef sen_obj *sen_column_create(sen_ctx *ctx, sen_obj *table,
                                  char *name, unsigned name_size,
                                  char *path, sen_obj_flags flags,
                                  sen_obj *type)
  cdef sen_obj *sen_column_open(sen_ctx *ctx, sen_obj *table,
                                char *name, unsigned name_size,
                                char *path, sen_obj *type)
  cdef sen_rc sen_column_index_update(sen_ctx *ctx, sen_obj *column,
                                      sen_id id, unsigned int section,
                                      sen_obj *oldvalue, sen_obj *newvalue)
  cdef sen_obj *sen_column_table(sen_ctx *ctx, sen_obj *column)

  ctypedef enum sen_info_type:
    SEN_INFO_ENCODING = 0
    SEN_INFO_SOURCE
    SEN_INFO_ELEMENT_SIZE
    SEN_INFO_CURR_MAX
    SEN_INFO_MAX_ELEMENT_SIZE
    SEN_INFO_SEG_SIZE
    SEN_INFO_CHUNK_SIZE
    SEN_INFO_INITIAL_N_SEGMENTS
    SEN_INFO_MAX_SECTION
    SEN_INFO_HOOK_LOCAL_DATA
    SEN_INFO_ELEMENT_A
    SEN_INFO_ELEMENT_CHUNK
    SEN_INFO_ELEMENT_CHUNK_SIZE
    SEN_INFO_ELEMENT_BUFFER_FREE
    SEN_INFO_ELEMENT_NTERMS
    SEN_INFO_ELEMENT_NTERMS_VOID
    SEN_INFO_ELEMENT_SIZE_IN_CHUNK
    SEN_INFO_ELEMENT_POS_IN_CHUNK
    SEN_INFO_ELEMENT_SIZE_IN_BUFFER
    SEN_INFO_ELEMENT_POS_IN_BUFFER
    SEN_INFO_ELEMENT_ESTIMATE_SIZE

  cdef sen_rc sen_obj_get_info(sen_ctx *ctx, sen_obj *obj,
                               sen_info_type type, sen_obj *valuebuf)
  cdef sen_rc sen_obj_set_info(sen_ctx *ctx, sen_obj *obj,
                               sen_info_type type, sen_obj *value)
  cdef sen_rc sen_obj_get_element_info(sen_ctx *ctx, sen_obj *obj,
                                       sen_id id,
                                       sen_info_type type, sen_obj *value)
  cdef sen_rc sen_obj_set_element_info(sen_ctx *ctx, sen_obj *obj,
                                       sen_id id,
                                       sen_info_type type, sen_obj *value)
  cdef int sen_obj_get_value(sen_ctx *ctx, sen_obj *obj, sen_id id,
                             sen_obj *value)
  cdef sen_rc sen_obj_set_value(sen_ctx *ctx, sen_obj *obj, sen_id id,
                                sen_obj *value, sen_update_type type)
  cdef sen_rc sen_obj_remove(sen_ctx *ctx, char *path)
  cdef sen_rc sen_obj_rename(sen_ctx *ctx, char *old_path, char *new_path)
  cdef sen_rc sen_obj_close(sen_ctx *ctx, sen_obj *obj)
  cdef char *sen_obj_path(sen_ctx *ctx, sen_obj *obj)
  cdef char *sen_obj_name(sen_ctx *ctx, sen_obj *obj)
  cdef int sen_obj_defrag(sen_ctx *ctx, sen_obj *obj, int threshold)
  cdef int sen_obj_expire(sen_ctx *ctx, sen_obj *obj, int threshold)
  cdef int sen_obj_check(sen_ctx *ctx, sen_obj *obj)
  cdef sen_rc sen_obj_lock(sen_ctx *ctx, sen_obj *obj, int timeout)
  cdef sen_rc sen_obj_unlock(sen_ctx *ctx, sen_obj *obj)
  cdef sen_rc sen_obj_clear_lock(sen_ctx *ctx, sen_obj *obj)
  cdef sen_obj *sen_obj_space(sen_ctx *ctx, sen_obj *obj)

  ctypedef struct sen_search_optarg:
    sen_search_flags flags
    int similarity_threshold
    int max_interval
    int *weight_vector
    int vector_size
    sen_obj *proc
    int max_size

  cdef sen_rc sen_obj_search(sen_ctx *ctx, sen_obj *obj, sen_obj *query,
                             sen_obj *res, sen_sel_operator op,
                             sen_search_optarg *optarg)
  cdef int sen_proc_call_next(sen_ctx *ctx, sen_obj *exec_info,
                              sen_obj *input, # cannot use 'in' on Python
                              sen_obj *out)
  cdef void *sen_proc_get_ctx_local_data(sen_ctx *ctx, sen_obj *exec_info)
  cdef void *sen_proc_get_hook_local_data(sen_ctx *ctx, sen_obj *exec_info)

  ctypedef enum sen_hook_entry:
    SEN_HOOK_SET = 0
    SEN_HOOK_GET
    SEN_HOOK_INSERT
    SEN_HOOK_DELETE

  cdef sen_rc sen_obj_add_hook(sen_ctx *ctx, sen_obj *obj,
                               sen_hook_entry entry,
                               int offset, sen_obj *proc, sen_obj *data)
  cdef sen_obj *sen_obj_get_hook(sen_ctx *ctx, sen_obj *obj,
                                 sen_hook_entry entry,
                                 int offset, sen_obj *data)
  cdef sen_rc sen_obj_delete_hook(sen_ctx *ctx, sen_obj *obj,
                                  sen_hook_entry entry, int offset)

  # define consts
  cdef enum:
    SEN_ID_NIL
    SEN_ID_MAX

    SEN_TABLE_MAX_KEY_SIZE

    SEN_OBJ_TABLE_HASH_KEY
    SEN_OBJ_TABLE_PAT_KEY
    SEN_OBJ_TABLE_NO_KEY
    SEN_OBJ_TABLE_ALIAS
    SEN_OBJ_TABLE_SUBSET

    SEN_OBJ_TOKEN_MASK
    SEN_OBJ_TOKEN_MECAB
    SEN_OBJ_TOKEN_NGRAM
    SEN_OBJ_TOKEN_DELIMITED
    SEN_OBJ_TOKEN_USER_DEFINED

    SEN_OBJ_KEY_WITH_SIS
    SEN_OBJ_KEY_NORMALIZE
    SEN_OBJ_KEY_SPLIT_ALPHA
    SEN_OBJ_KEY_SPLIT_DIGIT
    SEN_OBJ_KEY_SPLIT_SYMBOL

    SEN_OBJ_COLUMN_TYPE_MASK
    SEN_OBJ_COLUMN_SCALAR
    SEN_OBJ_COLUMN_ARRAY
    SEN_OBJ_COLUMN_INDEX

    SEN_OBJ_COMPRESS_MASK
    SEN_OBJ_COMPRESS_NONE
    SEN_OBJ_COMPRESS_ZLIB
    SEN_OBJ_COMPRESS_LZO

    SEN_OBJ_WITH_SECTION
    SEN_OBJ_NO_SECTION
    SEN_OBJ_WITH_SCORE
    SEN_OBJ_NO_SCORE
    SEN_OBJ_WITH_POSITION
    SEN_OBJ_NO_POSITION

    SEN_OBJ_WITH_SUBREC
    SEN_OBJ_NO_SUBREC

    SEN_OBJ_KEY_VAR_SIZE

    SEN_OBJ_INDEX_SCALAR

    SEN_OBJ_TEMPORARY
    SEN_OBJ_PERSISTENT

    SEN_VOID
    SEN_BULK
    SEN_VECTOR
    SEN_VERSES

    SEN_SEARCH_EXACT
    SEN_SEARCH_LCP
    SEN_SEARCH_SUFFIX
    SEN_SEARCH_PREFIX
    SEN_SEARCH_PARTIAL
    SEN_SEARCH_NEAR
    SEN_SEARCH_NEAR2
    SEN_SEARCH_SIMILAR
    SEN_SEARCH_TERM_EXTRACT

    SEN_TABLE_ADD
    SEN_TABLE_ADDED

    SEN_CURSOR_DESCENDING
    SEN_CURSOR_ASCENDING
    SEN_CURSOR_GE
    SEN_CURSOR_GT
    SEN_CURSOR_LE
    SEN_CURSOR_LT

    SEN_TABLE_SORT_DESC
    SEN_TABLE_SORT_ASC
    SEN_TABLE_SORT_BY_KEY
    SEN_TABLE_SORT_BY_VALUE
    SEN_TABLE_SORT_RES_ID
    SEN_TABLE_SORT_RES_KEY
    SEN_TABLE_SORT_AS_BIN
    SEN_TABLE_SORT_AS_NUMBER
    SEN_TABLE_SORT_AS_SIGNED
    SEN_TABLE_SORT_AS_UNSIGNED
    SEN_TABLE_SORT_AS_INT32
    SEN_TABLE_SORT_AS_INT64
    SEN_TABLE_SORT_NO_PROC
    SEN_TABLE_SORT_WITH_PROC

    SEN_TABLE_GROUP_BY_KEY
    SEN_TABLE_GROUP_BY_VALUE
    SEN_TABLE_GROUP_BY_SLOT_VALUE
    SEN_TABLE_GROUP_FILTER_PREFIX
    SEN_TABLE_GROUP_FILTER_SUFFIX
    SEN_TABLE_GROUP_CALC_COUNT
    SEN_TABLE_GROUP_CALC_MAX
    SEN_TABLE_GROUP_CALC_MIN
    SEN_TABLE_GROUP_CALC_SUM
    SEN_TABLE_GROUP_CALC_AVG
    SEN_TABLE_GROUP_CUMULATIVE

    SEN_PROC_MAX_ARGS

  # define macro functions
  cdef bint CTX_CLOSED "SEN_CTX_CLOSED" (sen_ctx *t)

cdef extern from "theheader.h":
  cdef object NEW_CLASS "PY_NEW" (object t)

# sen_obj
cdef class Object:
  cdef sen_obj *obj # if NULL, closed
  cdef sen_ctx *ctx

  def __init__(self):
    raise TypeError('This class cannot be instantiated from Python')

  def close(self):
    if self.obj == NULL:
      return
    rc = sen_obj_close(self.ctx, self.obj)
    if rc != sen_success:
      raise IOError('sen_obj_close failed')
    self.obj = NULL

  def __del__(self):
    self.close(self)

  # lookup is only allowed for space object
  def space_lookup(self, object name):
    cdef sen_obj *obj
    obj = sen_space_lookup(self.ctx, self.obj, name, PyString_Size(name))
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # type_create is only allowed for space object
  def type_create(self, object name, int flags, unsigned int size):
    cdef sen_obj *obj
    obj = sen_type_create(self.ctx, self.obj, name, PyString_Size(name),
                          flags, size)
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # TODO: implement!
  # ctypedef sen_rc sen_proc_func(sen_ctx *ctx, sen_proc_ctx *pctx,
  #                               int argc, sen_proc_data *argv)
  # ctypedef sen_rc sen_proc_init_func(sen_ctx *ctx, sen_obj *space,
  #                                    char *path)
  # cdef sen_obj *sen_proc_create(sen_ctx *ctx, sen_obj *space,
  #                               char *name, unsigned name_size,
  #                               char *path, sen_proc_type type,
  #                               sen_proc_func *init,
  #                               sen_proc_func *next,
  #                               sen_proc_func *fin)

  # table_create is only allowed for space object
  def table_create(self, object name, object path, int flags,
                   Object key_type, unsigned value_size, int encoding):
    cdef char *cpath
    cdef sen_obj *obj

    if path == None:
      cpath = NULL
    else:
      cpath = <char *>path

    obj = sen_table_create(self.ctx, self.obj, name, PyString_Size(name),
                           cpath, flags, key_type.obj, value_size,
                           <sen_encoding>encoding)

    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # table_open is only allowed for space object
  def table_open(self, object name, object path):
    cdef sen_obj *obj
    obj = sen_table_open(self.ctx, self.obj, name, PyString_Size(name),
                         path)
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # table_lookup is only allowed for table object
  def table_lookup(self, object key, sen_search_flags _flags):
    cdef sen_search_flags flags
    cdef sen_id id
    flags = _flags
    id = sen_table_lookup(self.ctx, self.obj, <char *>key, PyString_Size(key),
                          &flags)
    if (flags & SEN_TABLE_ADD) and not (flags & SEN_TABLE_ADDED):
      raise ValueError
    return id

  # table_add is only allowed for table object
  def table_add(self):
    cdef sen_id id
    id = sen_table_add(self.ctx, self.obj)
    return id

  # table_get_key is only allowed for table object
  def table_get_key(self, sen_id id):
    cdef char buf[SEN_TABLE_MAX_KEY_SIZE]
    cdef int key_size
    key_size = sen_table_get_key(self.ctx, self.obj, id, buf,
                                 SEN_TABLE_MAX_KEY_SIZE)
    if key_size == 0:
      return None
    return PyString_FromStringAndSize(buf, key_size)

  # table_delete is only allowed for table object
  def table_delete(self, object key):
    cdef sen_rc rc
    rc = sen_table_delete(self.ctx, self.obj,
                          <char *>key, PyString_Size(key))
    if rc != sen_success:
      raise ValueError
    return rc

  # table_delete_by_id is only allowed for table object
  def table_delete_by_id(self, sen_id id):
    cdef sen_rc rc
    rc = sen_table_delete_by_id(self.ctx, self.obj, id)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_truncate is only allowed for table object
  def table_truncate(self):
    cdef sen_rc rc
    rc = sen_table_truncate(self.ctx, self.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_cursor_open is only allowed for table object
  def table_cursor_open(self, object min, object max, int flags):
    cdef sen_table_cursor *cur
    cur = sen_table_cursor_open(self.ctx, self.obj,
                                <char *>min, PyString_Size(min),
                                <char *>max, PyString_Size(max),
                                flags)
    if cur == NULL:
      raise ValueError
    return _object_factory(self.ctx, cur) # sen_table_cursor == sen_obj

  # table_cursor_close is only allowed for table cursor object
  def table_cursor_close(self):
    cdef sen_rc rc
    rc = sen_table_cursor_close(self.ctx, self.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_cursor_next is only allowed for table cursor object
  def table_cursor_next(self):
    cdef sen_id id
    id = sen_table_cursor_next(self.ctx, self.obj)
    return id

  # table_cursor_get_key is only allowed for table cursor object
  def table_cursor_get_key(self):
    cdef char *key
    cdef int key_size
    key_size = sen_table_cursor_get_key(self.ctx, self.obj, <void **>&key)
    if key_size == 0:
      raise ValueError
    return PyString_FromStringAndSize(key, key_size)

  # table_cursor_get_value is only allowed for table cursor object
  def table_cursor_get_value(self):
    cdef char *value
    cdef int value_size
    value_size = sen_table_cursor_get_value(self.ctx, self.obj,
                                            <void **>&value)
    if value_size == 0:
      raise ValueError
    return PyString_FromStringAndSize(value, value_size)

  # table_cursor_set_value is only allowed for table cursor object
  def table_cursor_set_value(self, object value, int type):
    cdef sen_rc rc
    rc = sen_table_cursor_set_value(self.ctx, self.obj,
                                    <char *>value, <sen_update_type>type)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_cursor_delete is only allowed for table cursor object
  def table_cursor_delete(self):
    cdef sen_rc rc
    rc = sen_table_cursor_delete(self.ctx, self.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  # TODO: implement
  # cdef int sen_table_sort(sen_ctx *ctx, sen_obj *table, int limit,
  #                         sen_obj *result, sen_table_sort_optarg *optarg)

  # table_group is only allowed for table object
  def table_group(self, object args):
    if not PySequence_Check(args):
      return None
    cdef sen_rc rc
    cdef sen_table_group_arg *cargs
    cargs = <sen_table_group_arg *>malloc(
              sizeof(sen_table_group_arg) * len(args))
    for i in range(args):
      # TODO: implement!
      pass

    rc = sen_table_group(self.ctx, self.obj, len(args), cargs)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_setoperation is only allowed for table object
  def table_setoperation(self, Object obj, Object res, int op):
    cdef sen_rc rc
    rc = sen_table_setoperation(self.ctx, self.obj, obj.obj, res.obj,
                                <sen_sel_operator>res)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_difference is only allowed for table object
  def table_difference(self, Object obj, Object res1, Object res2):
    cdef sen_rc rc
    rc = sen_table_difference(self.ctx, self.obj, obj.obj,
                              res1.obj, res2.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  # table_column is only allowed for table object
  def table_column(self, object name):
    cdef sen_obj *obj
    obj = sen_table_column(self.ctx, self.obj,
                           name, PyString_Size(name))
    if obj == NULL:
      raise ValueError
    # FIXME: don't free!!!!!!
    return _object_factory(self.ctx, obj)

  # column_create is only allowed for table object
  def column_create(self, object name, object path, int flags,
                    Object type):
    cdef char *cpath
    cdef sen_obj *obj

    if path == None:
      cpath = NULL
    else:
      cpath = <char *>path

    obj = sen_column_create(self.ctx, self.obj,
                            name, PyString_Size(name),
                            cpath, <sen_obj_flags>flags, type.obj)
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # column_open is only allowed for table object
  def column_open(self, object name, char *path, int flags, Object type):
    cdef sen_obj *obj

    obj = sen_column_open(self.ctx, self.obj,
                          name, PyString_Size(name),
                          path, type.obj)
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  # column_index_update is only allowed for column object
  def column_index_update(self, sen_id id, unsigned int section,
                          Object oldvalue, Object newvalue):
    cdef sen_rc rc
    rc = sen_column_index_update(self.ctx, self.obj, id, section,
                                 oldvalue.obj, newvalue.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  # column_close is only allowed for column object
  def column_table(self):
    cdef sen_obj *obj
    obj = sen_column_table(self.ctx, self.obj)
    if obj == NULL:
      raise ValueError
    return _object_factory(self.ctx, obj)

  def get_info(self, sen_info_type type):
    cdef sen_rc rc
    cdef sen_obj *obj
    obj = <sen_obj *>malloc(sizeof(sen_obj))
    if obj == NULL:
      raise ValueError
    rc = sen_obj_get_info(self.ctx, self.obj, type, obj)
    if rc != sen_success:
      free(obj)
      raise ValueError
    return (rc, _object_factory(self.ctx, obj))

  def set_info(self, int type, Object value):
    cdef sen_rc rc
    rc = sen_obj_set_info(self.ctx, self.obj,
                          <sen_info_type>type, value.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  def get_element_info(self, sen_id id, int type):
    cdef sen_rc rc
    cdef sen_obj *obj
    obj = <sen_obj *>malloc(sizeof(sen_obj))
    if obj == NULL:
      raise ValueError
    rc = sen_obj_get_element_info(self.ctx, self.obj,
                                  id, <sen_info_type>type, obj)
    if rc != sen_success:
      free(obj)
      raise ValueError
    return (rc, _object_factory(self.ctx, obj))

  def set_element_info(self, sen_id id, int type, Object value):
    cdef sen_rc rc
    rc = sen_obj_set_element_info(self.ctx, self.obj, id,
                                  <sen_info_type>type, value.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  def get_value(self, sen_id id):
    # TODO: what is value type?
    cdef sen_obj *value
    cdef int value_len
    value = <sen_obj *>malloc(sizeof(sen_obj))
    if value == NULL:
      raise ValueError
    value_len = sen_obj_get_value(self.ctx, self.obj, id, value)
    if value_len == 0:
      free(value)
      raise ValueError
    return (value_len, _object_factory(self.ctx, value))

  def set_value(self, sen_id id, Object value, int type):
    cdef sen_rc rc
    rc = sen_obj_set_value(self.ctx, self.obj,
                           id, value.obj, <sen_update_type>type)
    if rc != sen_success:
      raise ValueError
    return rc

  def remove(self, Object path):
    cdef sen_rc rc
    rc = sen_obj_remove(self.ctx, path)
    if rc != sen_success:
      raise ValueError
    return rc

  def rename(self, Object old_path, Object new_path):
    cdef sen_rc rc
    rc = sen_obj_rename(self.ctx, old_path, new_path)
    if rc != sen_success:
      raise ValueError
    return rc

  def path(self):
    cdef char *path
    path = <char *>sen_obj_path(self.ctx, self.obj)
    if path == NULL:
      raise ValueError
    return PyString_FromString(path)

  def name(self):
    cdef char *name
    name = <char *>sen_obj_name(self.ctx, self.obj)
    if name == NULL:
      raise ValueError
    return PyString_FromString(name)

  def defrag(self, int threshold):
    cdef int v # TODO: rename
    v = sen_obj_defrag(self.ctx, self.obj, threshold)
    return v # TODO: check return value

  def expire(self, int threshold):
    cdef int v # TODO: rename
    v = sen_obj_expire(self.ctx, self.obj, threshold)
    return v # TODO: check return value

  def check(self):
    cdef int v # TODO: rename
    v = sen_obj_check(self.ctx, self.obj)

  def lock(self, int timeout):
    cdef sen_rc rc
    rc = sen_obj_lock(self.ctx, self.obj, timeout)
    if rc != sen_success:
      raise ValueError
    return rc

  def unlock(self):
    cdef sen_rc rc
    rc = sen_obj_unlock(self.ctx, self.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  def clear_lock(self):
    cdef sen_rc rc
    rc = sen_obj_clear_lock(self.ctx, self.obj)
    if rc != sen_success:
      raise ValueError
    return rc

  def space(self):
    cdef sen_obj *obj
    obj = sen_obj_space(self.ctx, self.obj)
    if obj == NULL:
      raise ValueError
    # FIXME: don't free!!!!!!
    return _object_factory(self.ctx, obj)

  # TODO: implement
  # cdef sen_rc sen_obj_search(sen_ctx *ctx, sen_obj *obj, sen_obj *query,
  #                            sen_obj *res, sen_sel_operator op,
  #                            sen_search_optarg *optarg)

  # cdef int sen_proc_call_next(sen_ctx *ctx, sen_obj *exec_info,
  #                             sen_obj *input, # cannot use in on Python
  #                             sen_obj *out)

  # cdef void *sen_proc_get_ctx_local_data(sen_ctx *ctx, sen_obj *exec_info)
  # cdef void *sen_proc_get_hook_local_data(sen_ctx *ctx, sen_obj *exec_info)

  # cdef sen_rc sen_obj_add_hook(sen_ctx *ctx, sen_obj *obj,
  #                              sen_hook_entry entry,
  #                              int offset, sen_obj *proc, sen_obj *data)
  # cdef sen_obj *sen_obj_get_hook(sen_ctx *ctx, sen_obj *obj,
  #                                sen_hook_entry entry,
  #                                int offset, sen_obj *data)
  # cdef sen_rc sen_obj_delete_hook(sen_ctx *ctx, sen_obj *obj,
  #                                 sen_hook_entry entry, int offset)

cdef Object _object_factory(sen_ctx *ctx, sen_obj *obj):
  cdef Object o
  o = NEW_CLASS(Object)
  o.ctx = ctx
  o.obj = obj
  return o

# Ctx
cdef class Ctx:
  cdef sen_ctx ctx

  def __init__(self, int flags, int encoding):
    rc = sen_ctx_init(&self.ctx, flags, <sen_encoding>encoding)
    if rc != sen_success:
      raise IOError('sen_ctx_init failed')

  def fin(self):
    rc = sen_ctx_fin(&self.ctx)
    if rc != sen_success:
      raise IOError('sen_ctx_fin failed')

  def __del__(self):
    self.fin(self)

  def space_create(self, object path, int encoding):
    cdef char *cpath
    cdef sen_obj *space

    if path == None:
      cpath = NULL
    else:
      cpath = <char *>path

    space = sen_space_create(&self.ctx, cpath, <sen_encoding>encoding)

    if space == NULL:
      raise IOError('sen_space_created failed')
    return _object_factory(&self.ctx, space)

  def space_open(self, char *path):
    cdef sen_obj *space
    space = sen_space_open(&self.ctx, path)
    if space == NULL:
      raise IOError('sen_space_open failed')
    return _object_factory(&self.ctx, space)

def py_sen_fin():
  sen_fin()

atexit.register(py_sen_fin)
sen_init()

# globals
# FIXME: access sen_type_any
# type_any = sen_type_any

# consts
# consts from enums
ID_NIL = SEN_ID_NIL
ID_MAX = SEN_ID_MAX

TABLE_MAX_KEY_SIZE = SEN_TABLE_MAX_KEY_SIZE

RC_SUCCESS = sen_success
RC_MEMORY_EXHAUSTED = sen_memory_exhausted
RC_INVALID_FORMAT = sen_invalid_format
RC_FILE_OPERATION_ERROR = sen_file_operation_error
RC_INVALID_ARGUMENT = sen_invalid_argument
RC_OTHER_ERROR = sen_other_error
RC_EXTERNAL_ERROR = sen_external_error
RC_INTERNAL_ERROR = sen_internal_error
RC_ABNORMAL_ERROR = sen_abnormal_error
RC_END_OF_DATA = sen_end_of_data

ENC_DEFAULT = sen_enc_default
ENC_NONE = sen_enc_none
ENC_EUC_JP = sen_enc_euc_jp
ENC_UTF8 = sen_enc_utf8
ENC_SJIS = sen_enc_sjis
ENC_LATIN1 = sen_enc_latin1
ENC_KOI8R = sen_enc_koi8r

SEL_OR = sen_sel_or
SEL_AND = sen_sel_and
SEL_BUT = sen_sel_but
SEL_ADJUST = sen_sel_adjust

PROC_HOOK = SEN_PROC_HOOK
PROC_RECALC = SEN_PROC_RECALC
PROC_SCORE = SEN_PROC_SCORE
PROC_COMPARE = SEN_PROC_COMPARE
PROC_GROUP = SEN_PROC_GROUP

UPD_SET = SEN_UPD_SET
UPD_ADD = SEN_UPD_ADD
UPD_CAS = SEN_UPD_CAS
UPD_APPEND = SEN_UPD_APPEND
UPD_PREPEND = SEN_UPD_PREPEND

INFO_ENCODING = SEN_INFO_ENCODING
INFO_SOURCE = SEN_INFO_SOURCE
INFO_ELEMENT_SIZE = SEN_INFO_ELEMENT_SIZE
INFO_CURR_MAX = SEN_INFO_CURR_MAX
INFO_MAX_ELEMENT_SIZE = SEN_INFO_MAX_ELEMENT_SIZE
INFO_SEG_SIZE = SEN_INFO_SEG_SIZE
INFO_CHUNK_SIZE = SEN_INFO_CHUNK_SIZE
INFO_INITIAL_N_SEGMENTS = SEN_INFO_INITIAL_N_SEGMENTS
INFO_MAX_SECTION = SEN_INFO_MAX_SECTION
INFO_HOOK_LOCAL_DATA = SEN_INFO_HOOK_LOCAL_DATA
INFO_ELEMENT_A = SEN_INFO_ELEMENT_A
INFO_ELEMENT_CHUNK = SEN_INFO_ELEMENT_CHUNK
INFO_ELEMENT_CHUNK_SIZE = SEN_INFO_ELEMENT_CHUNK_SIZE
INFO_ELEMENT_BUFFER_FREE = SEN_INFO_ELEMENT_BUFFER_FREE
INFO_ELEMENT_NTERMS = SEN_INFO_ELEMENT_NTERMS
INFO_ELEMENT_NTERMS_VOID = SEN_INFO_ELEMENT_NTERMS_VOID
INFO_ELEMENT_SIZE_IN_CHUNK = SEN_INFO_ELEMENT_SIZE_IN_CHUNK
INFO_ELEMENT_POS_IN_CHUNK = SEN_INFO_ELEMENT_POS_IN_CHUNK
INFO_ELEMENT_SIZE_IN_BUFFER = SEN_INFO_ELEMENT_SIZE_IN_BUFFER
INFO_ELEMENT_POS_IN_BUFFER = SEN_INFO_ELEMENT_POS_IN_BUFFER
INFO_ELEMENT_ESTIMATE_SIZE = SEN_INFO_ELEMENT_ESTIMATE_SIZE

HOOK_SET = SEN_HOOK_SET
HOOK_GET = SEN_HOOK_GET
HOOK_INSERT = SEN_HOOK_INSERT
HOOK_DELETE = SEN_HOOK_DELETE

# consts from define
#OBJ_TABLE_TYPE_MASK        = SEN_OBJ_TABLE_TYPE_MASK
OBJ_TABLE_HASH_KEY         = SEN_OBJ_TABLE_HASH_KEY
OBJ_TABLE_PAT_KEY          = SEN_OBJ_TABLE_PAT_KEY
OBJ_TABLE_NO_KEY           = SEN_OBJ_TABLE_NO_KEY
OBJ_TABLE_ALIAS            = SEN_OBJ_TABLE_ALIAS
OBJ_TABLE_SUBSET           = SEN_OBJ_TABLE_SUBSET

OBJ_TOKEN_MASK             = SEN_OBJ_TOKEN_MASK
OBJ_TOKEN_MECAB            = SEN_OBJ_TOKEN_MECAB
OBJ_TOKEN_NGRAM            = SEN_OBJ_TOKEN_NGRAM
OBJ_TOKEN_DELIMITED        = SEN_OBJ_TOKEN_DELIMITED
OBJ_TOKEN_USER_DEFINED     = SEN_OBJ_TOKEN_USER_DEFINED

OBJ_KEY_WITH_SIS           = SEN_OBJ_KEY_WITH_SIS
OBJ_KEY_NORMALIZE          = SEN_OBJ_KEY_NORMALIZE
OBJ_KEY_SPLIT_ALPHA        = SEN_OBJ_KEY_SPLIT_ALPHA
OBJ_KEY_SPLIT_DIGIT        = SEN_OBJ_KEY_SPLIT_DIGIT
OBJ_KEY_SPLIT_SYMBOL       = SEN_OBJ_KEY_SPLIT_SYMBOL

OBJ_COLUMN_TYPE_MASK       = SEN_OBJ_COLUMN_TYPE_MASK
OBJ_COLUMN_SCALAR          = SEN_OBJ_COLUMN_SCALAR
OBJ_COLUMN_ARRAY           = SEN_OBJ_COLUMN_ARRAY
OBJ_COLUMN_INDEX           = SEN_OBJ_COLUMN_INDEX

OBJ_COMPRESS_MASK          = SEN_OBJ_COMPRESS_MASK
OBJ_COMPRESS_NONE          = SEN_OBJ_COMPRESS_NONE
OBJ_COMPRESS_ZLIB          = SEN_OBJ_COMPRESS_ZLIB
OBJ_COMPRESS_LZO           = SEN_OBJ_COMPRESS_LZO

OBJ_WITH_SECTION           = SEN_OBJ_WITH_SECTION
OBJ_NO_SECTION             = SEN_OBJ_NO_SECTION
OBJ_WITH_SCORE             = SEN_OBJ_WITH_SCORE
OBJ_NO_SCORE               = SEN_OBJ_NO_SCORE
OBJ_WITH_POSITION          = SEN_OBJ_WITH_POSITION
OBJ_NO_POSITION            = SEN_OBJ_NO_POSITION

OBJ_WITH_SUBREC            = SEN_OBJ_WITH_SUBREC
OBJ_NO_SUBREC              = SEN_OBJ_NO_SUBREC

OBJ_KEY_VAR_SIZE           = SEN_OBJ_KEY_VAR_SIZE

OBJ_INDEX_SCALAR           = SEN_OBJ_INDEX_SCALAR

OBJ_TEMPORARY              = SEN_OBJ_TEMPORARY
OBJ_PERSISTENT             = SEN_OBJ_PERSISTENT

VOID                       = SEN_VOID
BULK                       = SEN_BULK
VECTOR                     = SEN_VECTOR
VERSES                     = SEN_VERSES

SEARCH_EXACT        = SEN_SEARCH_EXACT
SEARCH_LCP          = SEN_SEARCH_LCP
SEARCH_SUFFIX       = SEN_SEARCH_SUFFIX
SEARCH_PREFIX       = SEN_SEARCH_PREFIX
SEARCH_PARTIAL      = SEN_SEARCH_PARTIAL
SEARCH_NEAR         = SEN_SEARCH_NEAR
SEARCH_NEAR2        = SEN_SEARCH_NEAR2
SEARCH_SIMILAR      = SEN_SEARCH_SIMILAR
SEARCH_TERM_EXTRACT = SEN_SEARCH_TERM_EXTRACT

TABLE_ADD      = SEN_TABLE_ADD
TABLE_ADDED    = SEN_TABLE_ADDED

CURSOR_DESCENDING = SEN_CURSOR_DESCENDING
CURSOR_ASCENDING  = SEN_CURSOR_ASCENDING
CURSOR_GE = SEN_CURSOR_GE
CURSOR_GT = SEN_CURSOR_GT
CURSOR_LE = SEN_CURSOR_LE
CURSOR_LT = SEN_CURSOR_LT

TABLE_SORT_DESC        = SEN_TABLE_SORT_DESC
TABLE_SORT_ASC         = SEN_TABLE_SORT_ASC
TABLE_SORT_BY_KEY      = SEN_TABLE_SORT_BY_KEY
TABLE_SORT_BY_VALUE    = SEN_TABLE_SORT_BY_VALUE
TABLE_SORT_RES_ID      = SEN_TABLE_SORT_RES_ID
TABLE_SORT_RES_KEY     = SEN_TABLE_SORT_RES_KEY
TABLE_SORT_AS_BIN      = SEN_TABLE_SORT_AS_BIN
TABLE_SORT_AS_NUMBER   = SEN_TABLE_SORT_AS_NUMBER
TABLE_SORT_AS_SIGNED   = SEN_TABLE_SORT_AS_SIGNED
TABLE_SORT_AS_UNSIGNED = SEN_TABLE_SORT_AS_UNSIGNED
TABLE_SORT_AS_INT32    = SEN_TABLE_SORT_AS_INT32
TABLE_SORT_AS_INT64    = SEN_TABLE_SORT_AS_INT64
TABLE_SORT_NO_PROC     = SEN_TABLE_SORT_NO_PROC
TABLE_SORT_WITH_PROC   = SEN_TABLE_SORT_WITH_PROC

TABLE_GROUP_BY_KEY         = SEN_TABLE_GROUP_BY_KEY
TABLE_GROUP_BY_VALUE       = SEN_TABLE_GROUP_BY_VALUE
TABLE_GROUP_BY_SLOT_VALUE  = SEN_TABLE_GROUP_BY_SLOT_VALUE
TABLE_GROUP_FILTER_PREFIX  = SEN_TABLE_GROUP_FILTER_PREFIX
TABLE_GROUP_FILTER_SUFFIX  = SEN_TABLE_GROUP_FILTER_SUFFIX
TABLE_GROUP_CALC_COUNT     = SEN_TABLE_GROUP_CALC_COUNT
TABLE_GROUP_CALC_MAX       = SEN_TABLE_GROUP_CALC_MAX
TABLE_GROUP_CALC_MIN       = SEN_TABLE_GROUP_CALC_MIN
TABLE_GROUP_CALC_SUM       = SEN_TABLE_GROUP_CALC_SUM
TABLE_GROUP_CALC_AVG       = SEN_TABLE_GROUP_CALC_AVG
TABLE_GROUP_CUMULATIVE     = SEN_TABLE_GROUP_CUMULATIVE

PROC_MAX_ARGS = SEN_PROC_MAX_ARGS
