/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_DB_H
#define SEN_DB_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifdef	__cplusplus
extern "C" {
#endif

typedef struct {
  int score;
  int n_subrecs;
  int subrecs[1];
} sen_rset_recinfo;

typedef struct {
  sen_id rid;
  uint32_t sid;
  uint32_t pos;
} sen_rset_posinfo;

#define SEN_RSET_UTIL_BIT (0x80000000)

#define SEN_RSET_SCORE_SIZE (sizeof(int))

#define SEN_RSET_N_SUBRECS(ri) ((ri)->n_subrecs & ~SEN_RSET_UTIL_BIT)

#define SEN_RSET_SUBRECS_CMP(a,b,dir) (((a) - (b))*(dir) > 0)
#define SEN_RSET_SUBRECS_NTH(subrecs,size,n) \
  ((int *)((byte *)subrecs + n * (size + SEN_RSET_SCORE_SIZE)))
#define SEN_RSET_SUBRECS_COPY(subrecs,size,n,src) \
  (memcpy(SEN_RSET_SUBRECS_NTH(subrecs, size, n), src, size + SEN_RSET_SCORE_SIZE))



#define SEN_TABLE_DEFAULT_NGRAM_UNIT_SIZE 2

typedef struct _sen_space sen_space;
typedef struct _sen_proc sen_proc;
typedef struct _sen_table_delete_optarg sen_table_delete_optarg;

struct _sen_table_delete_optarg {
  int flags;
  int (*func)(sen_ctx *ctx, sen_obj *, sen_id, void *);
  void *func_arg;
};

sen_rc sen_space_close(sen_ctx *ctx, sen_space *s);

sen_obj *sen_space_keys(sen_obj *s);

sen_rc _sen_table_delete_by_id(sen_ctx *ctx, sen_obj *table, sen_id id,
                               sen_table_delete_optarg *optarg);

sen_id sen_table_at(sen_ctx *ctx, sen_obj *table, const void *key, int key_size,
                    void **value);
sen_id sen_table_get(sen_ctx *ctx, sen_obj *table, const void *key, int key_size,
                     void **value, sen_search_flags *flags);
sen_rc sen_table_get_info(sen_ctx *ctx, sen_obj *table, sen_obj_flags *flags,
                          sen_encoding *encoding, uint8_t *ngram_unit);
const char *_sen_table_key(sen_ctx *ctx, sen_obj *table, sen_id id, uint32_t *key_size);

sen_proc_data *sen_proc_ctx_get_local_data(sen_proc_ctx *pctx);

sen_rc sen_table_search(sen_ctx *ctx, sen_obj *table,
                        const void *key, uint32_t key_size,
                        sen_search_flags flags, sen_obj *res, sen_sel_operator op);

int sen_table_columns(sen_ctx *ctx, sen_obj *table, const char *name, unsigned name_size,
                      sen_obj *res);

sen_id sen_table_next(sen_ctx *ctx, sen_obj *table, sen_id id);

int sen_table_get_key2(sen_ctx *ctx, sen_obj *table, sen_id id, sen_obj *bulk);

sen_table_cursor *sen_table_cursor_open_by_id(sen_ctx *ctx, sen_obj *table,
                                              sen_id min, sen_id max, int flags);

void sen_table_add_subrec(sen_obj *table, sen_rset_recinfo *ri, int score,
                          sen_rset_posinfo *pi, int dir);


sen_obj *sen_obj_get_accessor(sen_ctx *ctx, sen_obj *obj,
                              const char *name, unsigned name_size);

sen_id sen_obj_get_range(sen_ctx *ctx, sen_obj *obj);

typedef struct _sen_hook sen_hook;

struct _sen_hook {
  sen_hook *next;
  sen_proc *proc;
  uint32_t hld_size;
};

typedef enum {
  PROC_INIT = 0,
  PROC_NEXT,
  PROC_FIN
} sen_proc_phase;

typedef struct {
  sen_obj_header header;
  sen_id id;
  sen_id range;  /* table: type of subrecords, column: type of values */
  sen_obj *space;
  sen_hook *hooks[5];
  void *source;
  uint32_t source_size;
  uint32_t max_n_subrecs;
  uint8_t subrec_size;
  uint8_t subrec_offset;
  uint8_t record_unit;
  uint8_t subrec_unit;
  sen_obj_flags flags;
} sen_obj_db;

#define SEN_OBJ_TMP_OBJECT 0x80000000

#define SEN_OBJ_DB_SET_TYPE(db_obj,obj_type) {\
  (db_obj)->obj.header.type = (obj_type);\
  (db_obj)->obj.header.impl_flags = 0;\
  (db_obj)->obj.header.flags = 0;\
  (db_obj)->obj.id = SEN_ID_NIL;\
  (db_obj)->obj.hooks[0] = NULL;\
  (db_obj)->obj.hooks[1] = NULL;\
  (db_obj)->obj.hooks[2] = NULL;\
  (db_obj)->obj.hooks[3] = NULL;\
  (db_obj)->obj.hooks[4] = NULL;\
  (db_obj)->obj.source = NULL;\
  (db_obj)->obj.source_size = 0;\
}

#define SEN_OBJ_DBP(obj) \
  (obj &&\
   (SEN_TYPE <= OBJ_DB(obj)->header.type) &&\
   (OBJ_DB(obj)->header.type <= SEN_COLUMN_INDEX))

#define SEN_OBJ_TABLEP(obj) \
  (obj &&\
   (SEN_TABLE_HASH_KEY <= OBJ_DB(obj)->header.type) &&\
   (OBJ_DB(obj)->header.type <= SEN_SPACE))

struct _sen_proc_ctx {
  sen_ctx *ctx;
  sen_obj *obj;
  sen_hook *hooks;
  sen_hook *currh;
  sen_proc_phase phase;
  unsigned short nargs;
  unsigned short offset;
  sen_proc_data data[16];
};

/* vector */

typedef struct _sen_vector sen_vector;

struct _sen_vector {
  sen_obj str;
  uint32_t *offsets;
  int n_entries;
};

const char *sen_vector_fetch(sen_ctx *ctx, sen_obj *vector, int i, unsigned int *size);
int sen_vector_delimit(sen_ctx *ctx, sen_obj *vector);
int sen_vector_size(sen_ctx *ctx, sen_obj *vector);

/* flag values used for sen_obj.header.impl_flags */

#define SEN_OBJ_ALLOCATED              (1L<<0) /* allocated by ctx */
#define SEN_OBJ_CUSTOM_NAME            (1L<<1) /* obj_db which has custom name */

#ifdef __cplusplus
}
#endif

#endif /* SEN_DB_H */
