/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_HASH_H
#define SEN_HASH_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifndef SEN_CTX_H
#include "ctx.h"
#endif /* SEN_CTX_H */

#ifndef SEN_DB_H
#include "db.h"
#endif /* SEN_DB_H */

#ifdef	__cplusplus
extern "C" {
#endif

/**** sen_tiny_array ****/

#define SEN_TINY_ARRAY_CLEAR      (1L<<0)
#define SEN_TINY_ARRAY_THREADSAFE (1L<<1)
#define SEN_TINY_ARRAY_USE_MALLOC (1L<<2)

#define SEN_TINY_ARRAY_W 0
#define SEN_TINY_ARRAY_R(i) (1<<((i)<<SEN_TINY_ARRAY_W))
#define SEN_TINY_ARRAY_S (SEN_TINY_ARRAY_R(1)-1)
#define SEN_TINY_ARRAY_N (32>>SEN_TINY_ARRAY_W)

typedef struct _sen_tiny_array sen_tiny_array;

struct _sen_tiny_array {
  sen_ctx *ctx;
  sen_id max;
  uint16_t element_size;
  uint16_t flags;
  sen_mutex lock;
  void *elements[SEN_TINY_ARRAY_N];
};

#define SEN_TINY_ARRAY_EACH(a,head,tail,key,val,block) do {\
  int _ei;\
  const sen_id h = (head);\
  const sen_id t = (tail);\
  for (_ei = 0, (key) = (h); _ei < SEN_TINY_ARRAY_N && (key) <= (t); _ei++) {\
    int _ej = SEN_TINY_ARRAY_S * SEN_TINY_ARRAY_R(_ei);\
    if (((val) = (a)->elements[_ei])) {\
      for (; _ej-- && (key) <= (t); (key)++, (val) = (void *)((byte *)(val) + (a)->element_size)) block\
    } else {\
      (key) += _ej;\
    }\
  }\
} while (0)

#define SEN_TINY_ARRAY_AT(a,id,e) do {\
  sen_id id_ = (id);\
  byte **e_;\
  size_t o_;\
  {\
    int l_, i_;\
    if (!id_) { e = NULL; break; }\
    SEN_BIT_SCAN_REV(id_, l_);\
    i_ = l_ >> SEN_TINY_ARRAY_W;\
    o_ = SEN_TINY_ARRAY_R(i_);\
    e_ = (byte **)&(a)->elements[i_];\
  }\
  if (!*e_) {\
    sen_ctx *ctx = (a)->ctx;\
    if ((a)->flags & SEN_TINY_ARRAY_THREADSAFE) { MUTEX_LOCK((a)->lock); }\
    if (!*e_) {\
      if ((a)->flags & SEN_TINY_ARRAY_USE_MALLOC) {\
        if ((a)->flags & SEN_TINY_ARRAY_CLEAR) {\
          *e_ = SEN_CALLOC(SEN_TINY_ARRAY_S * o_ * (a)->element_size);\
        } else {\
          *e_ = SEN_MALLOC(SEN_TINY_ARRAY_S * o_ * (a)->element_size);\
        }\
      } else {\
        *e_ = SEN_CTX_ALLOC(ctx, SEN_TINY_ARRAY_S * o_ * (a)->element_size, 0);\
      }\
    }\
    if ((a)->flags & SEN_TINY_ARRAY_THREADSAFE) { MUTEX_UNLOCK((a)->lock); }\
    if (!*e_) { e = NULL; break; }\
  }\
  if (id_ > (a)->max) { (a)->max = id_; }\
  (e) = (void *)(*e_ + (id_ - o_) * (a)->element_size);\
} while (0)

#define SEN_TINY_ARRAY_NEXT(a,e) SEN_TINY_ARRAY_AT((a), (a)->max + 1, e)

#define SEN_TINY_ARRAY_BIT_AT(array,offset,res) {\
  uint8_t *ptr_;\
  SEN_TINY_ARRAY_AT((array), ((offset) >> 3) + 1, ptr_);\
  res = ptr_ ? ((*ptr_ >> ((offset) & 7)) & 1) : 0;\
}

#define SEN_TINY_ARRAY_BIT_ON(array,offset) {\
  uint8_t *ptr_;\
  SEN_TINY_ARRAY_AT((array), ((offset) >> 3) + 1, ptr_);\
  if (ptr_) { *ptr_ |= (1 << ((offset) & 7)); }\
}

#define SEN_TINY_ARRAY_BIT_OFF(array,offset) {\
  uint8_t *ptr_;\
  SEN_TINY_ARRAY_AT((array), ((offset) >> 3) + 1, ptr_);\
  if (ptr_) { *ptr_ &= ~(1 << ((offset) & 7)); }\
}

#define SEN_TINY_ARRAY_BIT_FLIP(array,offset) {\
  uint8_t *ptr_;\
  SEN_TINY_ARRAY_AT((array), ((offset) >> 3) + 1, ptr_);\
  if (ptr_) { *ptr_ ^= (1 << ((offset) & 7)); }\
}

void sen_tiny_array_init(sen_tiny_array *a, sen_ctx *ctx, uint16_t element_size, uint16_t flags);
void sen_tiny_array_fin(sen_tiny_array *a);
void *sen_tiny_array_at(sen_tiny_array *a, sen_id id);
sen_id sen_tiny_array_id(sen_tiny_array *a, void *p);

/**** sen_array ****/

typedef struct _sen_array sen_array;

#define SEN_ARRAY_TINY (1L<<1)

typedef struct _sen_array_cursor sen_array_cursor;

struct _sen_array {
  sen_obj_db obj;
  sen_ctx *ctx;
  uint32_t value_size;
  uint32_t *n_garbages;
  uint32_t *n_entries;
  /* portions for io_array */
  sen_io *io;
  struct sen_array_header *header;
  uint32_t *lock;
  /* portions for tiny_array */
  uint32_t n_garbages_;
  uint32_t n_entries_;
  sen_id garbages;
  sen_tiny_array a;
  sen_tiny_array bitmap;
};

struct _sen_array_cursor {
  sen_obj_db obj;
  sen_id curr_rec;
  sen_array *array;
  sen_ctx *ctx;
  int dir;
  sen_id limit;
};

sen_array *sen_array_create(sen_ctx *ctx, const char *path,
                            uint32_t value_size, uint32_t flags);
sen_array *sen_array_open(sen_ctx *ctx, const char *path);
sen_rc sen_array_close(sen_ctx *ctx, sen_array *array);
sen_id sen_array_add(sen_ctx *ctx, sen_array *array, void **value);
int sen_array_get_value(sen_ctx *ctx, sen_array *array, sen_id id, void *valuebuf);
sen_rc sen_array_set_value(sen_ctx *ctx, sen_array *array, sen_id id,
                           void *value, int flags);
sen_array_cursor *sen_array_cursor_open(sen_ctx *ctx, sen_array *array,
                                        sen_id min, sen_id max, int flags);
sen_id sen_array_cursor_next(sen_ctx *ctx, sen_array_cursor *c);
int sen_array_cursor_get_value(sen_ctx *ctx, sen_array_cursor *c, void **value);
sen_rc sen_array_cursor_set_value(sen_ctx *ctx, sen_array_cursor *c,
                                  void *value, int flags);
sen_rc sen_array_cursor_delete(sen_ctx *ctx, sen_array_cursor *c,
                               sen_table_delete_optarg *optarg);
void sen_array_cursor_close(sen_ctx *ctx, sen_array_cursor *c);
sen_rc sen_array_delete_by_id(sen_ctx *ctx, sen_array *array, sen_id id,
                              sen_table_delete_optarg *optarg);

sen_id sen_array_next(sen_ctx *ctx, sen_array *array, sen_id id);

void *_sen_array_get_value(sen_ctx *ctx, sen_array *array, sen_id id);

#define SEN_ARRAY_EACH(array,head,tail,id,value,block) do {\
  sen_array_cursor *_sc = sen_array_cursor_open(ctx, array, head, tail, 0);\
  if (_sc) {\
    sen_id id;\
    while ((id = sen_array_cursor_next(ctx, _sc))) {\
      sen_array_cursor_get_value(ctx, _sc, (void **)(value)); \
      block\
    }\
    sen_array_cursor_close(ctx, _sc); \
  }\
} while (0)

#define SEN_ARRAY_SIZE(array) (*((array)->n_entries))

/**** sen_hash ****/

#define SEN_HASH_TINY (1L<<1)
#define SEN_HASH_MAX_KEY_SIZE SEN_TABLE_MAX_KEY_SIZE

typedef struct _sen_hash sen_hash;
typedef struct _sen_hash_cursor sen_hash_cursor;

struct _sen_hash {
  sen_obj_db obj;
  sen_ctx *ctx;
  uint32_t key_size;
  sen_encoding encoding;
  uint32_t value_size;
  uint32_t entry_size;
  uint32_t *n_garbages;
  uint32_t *n_entries;
  uint32_t *max_offset;
  uint8_t ngram_unit;
  /* portions for io_hash */
  sen_io *io;
  struct sen_hash_header *header;
  uint32_t *lock;
  // uint32_t nref;
  // unsigned int max_n_subrecs;
  // unsigned int record_size;
  // unsigned int subrec_size;
  // sen_rec_unit record_unit;
  // sen_rec_unit subrec_unit;
  // uint8_t arrayp;
  // sen_recordh *curr_rec;
  // sen_set_cursor *cursor;
  // int limit;
  // void *userdata;
  // sen_id subrec_id;
  /* portions for tiny_hash */
  uint32_t max_offset_;
  uint32_t n_garbages_;
  uint32_t n_entries_;
  sen_id *index;
  sen_id garbages;
  sen_tiny_array a;
  sen_tiny_array bitmap;
};

struct sen_hash_header {
  char idstr[16];
  uint32_t flags;
  sen_encoding encoding;
  uint32_t key_size;
  uint32_t curr_rec;
  int32_t curr_key;
  uint32_t idx_offset;
  uint32_t value_size;
  uint32_t entry_size;
  uint32_t max_offset;
  uint32_t n_entries;
  uint32_t n_garbages;
  uint32_t lock;
  uint8_t ngram_unit;
  uint32_t reserved[17];
  sen_id garbages[SEN_HASH_MAX_KEY_SIZE];
};

struct _sen_hash_cursor {
  sen_obj_db obj;
  sen_id curr_rec;
  sen_hash *hash;
  sen_ctx *ctx;
  int dir;
  sen_id limit;
};

/* deprecated */

#define SEN_TABLE_SORT_BY_KEY      0
#define SEN_TABLE_SORT_BY_ID       (1L<<1)
#define SEN_TABLE_SORT_BY_VALUE    (1L<<2)
#define SEN_TABLE_SORT_RES_ID      0
#define SEN_TABLE_SORT_RES_KEY     (1L<<3)
#define SEN_TABLE_SORT_AS_BIN      0
#define SEN_TABLE_SORT_AS_NUMBER   (1L<<4)
#define SEN_TABLE_SORT_AS_SIGNED   0
#define SEN_TABLE_SORT_AS_UNSIGNED (1L<<5)
#define SEN_TABLE_SORT_AS_INT32    0
#define SEN_TABLE_SORT_AS_INT64    (1L<<6)
#define SEN_TABLE_SORT_NO_PROC     0
#define SEN_TABLE_SORT_WITH_PROC   (1L<<7)

typedef struct _sen_table_sort_optarg sen_table_sort_optarg;

struct _sen_table_sort_optarg {
  sen_table_sort_flags flags;
  int (*compar)(sen_ctx *ctx,
                sen_obj *table1, void *target1, unsigned int target1_size,
                sen_obj *table2, void *target2, unsigned int target2_size,
                void *compare_arg);
  void *compar_arg;
  sen_obj *proc;
  int offset;
};


sen_hash *sen_hash_create(sen_ctx *ctx, const char *path, uint32_t key_size,
                          uint32_t value_size, uint32_t flags, sen_encoding encoding);

sen_hash *sen_hash_open(sen_ctx *ctx, const char *path);

sen_rc sen_hash_close(sen_ctx *ctx, sen_hash *hash);

sen_id sen_hash_lookup(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size,
                       void **value, sen_search_flags *flags);

int sen_hash_get_key(sen_ctx *ctx, sen_hash *hash, sen_id id, void *keybuf, int bufsize);
int sen_hash_get_key2(sen_ctx *ctx, sen_hash *hash, sen_id id, sen_obj *bulk);
int sen_hash_get_value(sen_ctx *ctx, sen_hash *hash, sen_id id, void *valuebuf);
sen_rc sen_hash_set_value(sen_ctx *ctx, sen_hash *hash, sen_id id, void *value,
                          int flags);

sen_rc sen_hash_delete_by_id(sen_ctx *ctx, sen_hash *hash, sen_id id,
                             sen_table_delete_optarg *optarg);
sen_rc sen_hash_delete(sen_ctx *ctx, sen_hash *hash, const void *key, uint32_t key_size,
                       sen_table_delete_optarg *optarg);

sen_hash_cursor *sen_hash_cursor_open(sen_ctx *ctx, sen_hash *hash,
                                      const void *min, uint32_t min_size,
                                      const void *max, uint32_t max_size, int flags);
sen_id sen_hash_cursor_next(sen_ctx *ctx, sen_hash_cursor *c);
void sen_hash_cursor_close(sen_ctx *ctx, sen_hash_cursor *c);

int sen_hash_cursor_get_key(sen_ctx *ctx, sen_hash_cursor *c, void **key);
int sen_hash_cursor_get_value(sen_ctx *ctx, sen_hash_cursor *c, void **value);
sen_rc sen_hash_cursor_set_value(sen_ctx *ctx, sen_hash_cursor *c,
                                 void *value, int flags);
sen_rc sen_hash_cursor_delete(sen_ctx *ctx, sen_hash_cursor *c,
                              sen_table_delete_optarg *optarg);
int sen_hash_sort(sen_ctx *ctx, sen_hash *hash, int limit,
                  sen_array *result, sen_table_sort_optarg *optarg);

int sen_hash_cursor_get_key_value(sen_ctx *ctx, sen_hash_cursor *c,
                                  void **key, uint32_t *key_size, void **value);
sen_id sen_hash_at(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size,
                   void **value);
sen_id sen_hash_get(sen_ctx *ctx, sen_hash *hash, const void *key, int key_size,
                    void **value, sen_search_flags *flags);

sen_rc sen_hash_lock(sen_ctx *ctx, sen_hash *hash, int timeout);
sen_rc sen_hash_unlock(sen_ctx *ctx, sen_hash *hash);
sen_rc sen_hash_clear_lock(sen_ctx *ctx, sen_hash *hash);

#define SEN_HASH_EACH(hash,id,key,key_size,value,block) do {\
  sen_hash_cursor *_sc = sen_hash_cursor_open(ctx, hash, NULL, 0, NULL, 0, 0);\
  if (_sc) {\
    sen_id id;\
    while ((id = sen_hash_cursor_next(ctx, _sc))) {\
      sen_hash_cursor_get_key_value(ctx, _sc, (void **)(key),\
                                    (key_size), (void **)(value));\
      block\
    }\
    sen_hash_cursor_close(ctx, _sc);\
  }\
} while (0)

#define SEN_HASH_SIZE(hash) (*((hash)->n_entries))

/* private */
int sen_rec_unit_size(sen_rec_unit unit, int rec_size);

const char * _sen_hash_key(sen_hash *hash, sen_id id, uint32_t *key_size);

int sen_hash_get_key_value(sen_ctx *ctx, sen_hash *hash, sen_id id,
                           void *keybuf, int bufsize, void *valuebuf);

void sen_rhash_add_subrec(sen_hash *s, sen_rset_recinfo *ri,
                          int score, void *body, int dir);

int _sen_hash_get_key_value(sen_ctx *ctx, sen_hash *hash, sen_id id,
                            void **key, void **value);

sen_id sen_hash_next(sen_ctx *ctx, sen_hash *hash, sen_id id);

/* only valid for hash tables, SEN_OBJ_KEY_VAR_SIZE && SEN_HASH_TINY */
const char *_sen_hash_strkey_by_val(void *v, uint16_t *size);

const char *sen_hash_get_value_(sen_ctx *ctx, sen_hash *hash, sen_id id, uint32_t *size);

#ifdef __cplusplus
}
#endif

#endif /* SEN_HASH_H */
