/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_II_H
#define SEN_II_H

/* "ii" is for inverted index */

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifndef SEN_HASH_H
#include "hash.h"
#endif /* SEN_HASH_H */

#ifndef SEN_IO_H
#include "io.h"
#endif /* SEN_IO_H */

#ifndef SEN_STORE_H
#include "store.h"
#endif /* SEN_STORE_H */

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct _sen_ii sen_ii;

struct _sen_ii {
  sen_obj_db obj;
  sen_io *seg;
  sen_io *chunk;
  sen_obj *lexicon;
  sen_obj_flags lflags;
  sen_encoding encoding;
  uint32_t n_elements;
  uint32_t max_n_elements;
  struct sen_ii_header *header;
};

struct sen_ii_header;

struct _sen_ii_pos {
  struct _sen_ii_pos *next;
  uint32_t pos;
};

struct _sen_ii_updspec {
  uint32_t rid;
  uint32_t sid;
  int32_t score;
  int32_t tf;                 /* number of postings successfully stored to index */
  int32_t atf;                /* actual number of postings */
  struct _sen_ii_pos *pos;
  struct _sen_ii_pos *tail;
  sen_vgram_vnode *vnodes;
};

typedef struct _sen_ii_updspec sen_ii_updspec;

sen_ii *sen_ii_create(sen_ctx *ctx, const char *path, sen_obj *lexicon, uint32_t flags);
sen_ii *sen_ii_open(sen_ctx *ctx, const char *path, sen_obj *lexicon);
sen_rc sen_ii_close(sen_ctx *ctx, sen_ii *ii);
sen_rc sen_ii_remove(sen_ctx *ctx, const char *path);
sen_rc sen_ii_info(sen_ctx *ctx, sen_ii *ii, uint64_t *seg_size, uint64_t *chunk_size);
sen_rc sen_ii_update_one(sen_ctx *ctx, sen_ii *ii, uint32_t key, sen_ii_updspec *u,
                         sen_hash *h);
sen_rc sen_ii_delete_one(sen_ctx *ctx, sen_ii *ii, uint32_t key, sen_ii_updspec *u,
                         sen_hash *h);
sen_ii_updspec *sen_ii_updspec_open(sen_ctx *ctx, uint32_t rid, uint32_t sid);
sen_rc sen_ii_updspec_close(sen_ctx *ctx, sen_ii_updspec *u);
sen_rc sen_ii_updspec_add(sen_ctx *ctx, sen_ii_updspec *u, int pos, int32_t weight);
int sen_ii_updspec_cmp(sen_ii_updspec *a, sen_ii_updspec *b);

uint32_t sen_ii_estimate_size(sen_ii *ii, uint32_t key);

void sen_ii_seg_expire(sen_ii *ii, int32_t threshold);
void sen_ii_expire(sen_ctx *ctx, sen_ii *ii);

typedef struct {
  sen_id rid;
  uint32_t sid;
  uint32_t pos;
  uint32_t tf;
  uint32_t score;
  uint32_t rest;
} sen_ii_posting;

typedef struct _sen_ii_cursor sen_ii_cursor;

sen_ii_cursor *sen_ii_cursor_open(sen_ctx *ctx, sen_ii *ii, sen_id tid,
                                  sen_id min, sen_id max, int nelements, int flags);
sen_ii_cursor *sen_ii_cursor_openv1(sen_ii *ii, uint32_t key);
sen_rc sen_ii_cursor_openv2(sen_ii_cursor **cursors, int ncursors);
sen_ii_posting *sen_ii_cursor_next(sen_ctx *ctx, sen_ii_cursor *c);
sen_ii_posting *sen_ii_cursor_next_pos(sen_ctx *ctx, sen_ii_cursor *c);
sen_rc sen_ii_cursor_close(sen_ctx *ctx, sen_ii_cursor *c);

uint32_t sen_ii_max_section(sen_ii *ii);

int sen_ii_check(sen_ii *ii);
const char *sen_ii_path(sen_ii *ii);
sen_obj *sen_ii_lexicon(sen_ii *ii);

sen_rc sen_ii_upd(sen_ctx *ctx, sen_ii *ii, sen_id rid, sen_vgram *vgram,
                   const char *oldvalue, unsigned int oldvalue_len,
                   const char *newvalue, unsigned int newvalue_len);
sen_rc sen_ii_update(sen_ctx *ctx, sen_ii *ii, sen_id rid, sen_vgram *vgram,
                      unsigned int section,
                      sen_values *oldvalues, sen_values *newvalues);
sen_rc sen_ii_column_update(sen_ctx *ctx, sen_ii *ii, sen_id id, unsigned int section,
                            sen_obj *oldvalue, sen_obj *newvalue);
sen_rc sen_ii_term_extract(sen_ctx *ctx, sen_ii *ii, const char *string,
                            unsigned int string_len, sen_hash *s,
                            sen_sel_operator op, sen_select_optarg *optarg);
sen_rc sen_ii_similar_search(sen_ctx *ctx, sen_ii *ii, const char *string, unsigned int string_len,
                              sen_hash *s, sen_sel_operator op, sen_select_optarg *optarg);
sen_rc sen_ii_select(sen_ctx *ctx, sen_ii *ii, const char *string, unsigned int string_len,
                     sen_hash *s, sen_sel_operator op, sen_select_optarg *optarg);
sen_rc sen_ii_sel(sen_ctx *ctx, sen_ii *ii, const char *string, unsigned int string_len,
                  sen_hash *s);

sen_rc sen_ii_query_select(sen_ctx *ctx, sen_ii *i, sen_query *q, sen_records *r, sen_sel_operator op);

sen_rc sen_query_search(sen_ctx *ctx, sen_ii *i, sen_query *q,
                        sen_hash *r, sen_sel_operator op);

#ifdef __cplusplus
}
#endif

#endif /* SEN_II_H */
