/* Copyright(C) 2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_PAT_H
#define SEN_PAT_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#include "db.h"
#include "hash.h"

#ifdef	__cplusplus
extern "C" {
#endif

#define SEN_PAT_MAX_KEY_SIZE                    SEN_TABLE_MAX_KEY_SIZE

typedef struct _sen_pat_scan_hit sen_pat_scan_hit;
typedef struct _sen_pat_cursor sen_pat_cursor;

struct _sen_pat {
  sen_obj_db obj;
  sen_io *io;
  struct sen_pat_header *header;
  sen_encoding encoding;
  uint32_t key_size;
  uint32_t value_size;
  uint8_t ngram_unit;
  uint32_t nref; // not in use
};

struct _sen_pat_scan_hit {
  sen_id id;
  unsigned int offset;
  unsigned int length;
};

typedef struct _sen_pat sen_pat;

struct _sen_pat_cursor_entry {
  sen_id id;
  uint16_t check;
};

typedef struct _sen_pat_cursor_entry sen_pat_cursor_entry;

struct _sen_pat_cursor {
  sen_obj_db obj;
  sen_id curr_rec;
  sen_pat *pat;
  sen_ctx *ctx;
  unsigned int size;
  unsigned int sp;
  sen_id limit;
  sen_pat_cursor_entry *ss;
  uint8_t curr_key[SEN_TABLE_MAX_KEY_SIZE];
};

sen_pat *sen_pat_create(sen_ctx *ctx, const char *path, uint32_t key_size,
                        uint32_t value_size, uint32_t flags, sen_encoding encoding);

sen_pat *sen_pat_open(sen_ctx *ctx, const char *path);

sen_rc sen_pat_close(sen_ctx *ctx, sen_pat *pat);

sen_rc sen_pat_remove(sen_ctx *ctx, const char *path);

sen_id sen_pat_lookup(sen_ctx *ctx, sen_pat *pat, const void *key, int key_size,
                       void **value, sen_search_flags *flags);

sen_rc sen_pat_prefix_search(sen_ctx *ctx, sen_pat *pat,
                             const void *key, uint32_t key_size, sen_hash *h);
sen_rc sen_pat_suffix_search(sen_ctx *ctx, sen_pat *pat,
                             const void *key, uint32_t key_size, sen_hash *h);
sen_id sen_pat_lcp_search(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size);

uint32_t sen_pat_size(sen_ctx *ctx, sen_pat *pat);

int sen_pat_get_key(sen_ctx *ctx, sen_pat *pat, sen_id id, void *keybuf, int bufsize);
int sen_pat_get_key2(sen_ctx *ctx, sen_pat *pat, sen_id id, sen_obj *bulk);
int sen_pat_get_value(sen_ctx *ctx, sen_pat *pat, sen_id id, void *valuebuf);
sen_rc sen_pat_set_value(sen_ctx *ctx, sen_pat *pat, sen_id id,
                         void *value, int flags);

sen_rc sen_pat_delete_by_id(sen_ctx *ctx, sen_pat *pat, sen_id id,
                            sen_table_delete_optarg *optarg);
sen_rc sen_pat_delete(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
                      sen_table_delete_optarg *optarg);
int sen_pat_delete_with_sis(sen_ctx *ctx, sen_pat *pat, sen_id id,
                      sen_table_delete_optarg *optarg);

int sen_pat_scan(sen_ctx *ctx, sen_pat *pat, const char *str, unsigned int str_len,
                 sen_pat_scan_hit *sh, unsigned int sh_size, const char **rest);

sen_pat_cursor *sen_pat_cursor_open(sen_ctx *ctx, sen_pat *pat,
                                    const void *min, uint32_t min_size,
                                    const void *max, uint32_t max_size, int flags);
sen_id sen_pat_cursor_next(sen_ctx *ctx, sen_pat_cursor *c);
void sen_pat_cursor_close(sen_ctx *ctx, sen_pat_cursor *c);

int sen_pat_cursor_get_key(sen_ctx *ctx, sen_pat_cursor *c, void **key);
int sen_pat_cursor_get_value(sen_ctx *ctx, sen_pat_cursor *c, void **value);
int sen_pat_cursor_get_key_value(sen_ctx *ctx, sen_pat_cursor *c,
                                 void **key, uint32_t *key_size, void **value);
sen_rc sen_pat_cursor_set_value(sen_ctx *ctx, sen_pat_cursor *c,
                                void *value, int flags);
sen_rc sen_pat_cursor_delete(sen_ctx *ctx, sen_pat_cursor *c,
                             sen_table_delete_optarg *optarg);

sen_id sen_pat_at(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
                  void **value);
sen_id sen_pat_get(sen_ctx *ctx, sen_pat *pat, const void *key, uint32_t key_size,
                   void **value, sen_search_flags *flags);
sen_id sen_pat_curr_id(sen_ctx *ctx, sen_pat *pat);

#define SEN_PAT_EACH(pat,id,key,key_size,value,block) do {\
  sen_pat_cursor *_sc = sen_pat_cursor_open(ctx, pat, NULL, 0, NULL, 0, 0);\
  if (_sc) {\
    sen_id id;\
    while ((id = sen_pat_cursor_next(ctx, _sc))) {\
      sen_pat_cursor_get_key_value(ctx, _sc, (void **)(key),\
                                    (key_size), (void **)(value));\
      block\
    }\
    sen_pat_cursor_close(ctx, _sc);\
  }\
} while (0)

/* private */
const char *_sen_pat_key(sen_pat *pat, sen_id id, uint32_t *key_size);
sen_id sen_pat_next(sen_ctx *ctx, sen_pat *pat, sen_id id);
const char *sen_pat_get_value_(sen_ctx *ctx, sen_pat *pat, sen_id id, uint32_t *size);

#ifdef __cplusplus
}
#endif

#endif /* SEN_PAT_H */
