/* Copyright(C) 2004-2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_QL_H
#define SEN_QL_H

#ifndef SEN_STORE_H
#include "store.h"
#endif /* SEN_STORE_H */

#ifndef SEN_COM_H
#include "com.h"
#endif /* SEN_COM_H */

#ifndef SEN_CTX_H
#include "ctx.h"
#endif /* SEN_CTX_H */

#ifndef _ISOC99_SOURCE
#define _ISOC99_SOURCE
#endif /* _ISOC99_SOURCE */
#include <math.h>

#ifdef __cplusplus
extern "C" {
#endif

/**** sen_cell ****/

/* flag values used for sen_cell.header.impl_flags */
/* bit 0~1 are reserved */
#define SEN_CELL_SYMBOL                (1L<<3) /* ql: registered in the symbol table */
#define SEN_CELL_PROMISE               (1L<<4) /* ql: promise object */
#define SEN_CELL_NATIVE                (1L<<5) /* ql: native function */
#define SEN_CELL_MARKED                (1L<<6) /* ql: used in garbage collection */
#define SEN_CELL_MARK2                 (1L<<7) /* ql: used in callback function */

/* type values used for sen_cell.header.type */
#define SEN_CELL_INT                   (0x70)
#define SEN_CELL_FLOAT                 (0x71)
#define SEN_CELL_STR                   (0x72)
#define SEN_CELL_OP                    (0x73)
#define SEN_CELL_OBJECT                (0x74)
#define SEN_CELL_PROC                  (0x75)
#define SEN_CELL_SYNTAX                (0x76)
#define SEN_CELL_TIME                  (0x77)
#define SEN_CELL_PSEUDO_COLUMN         (0x78)
#define SEN_CELL_LIST                  (0x80)
#define SEN_CELL_CLOSURE               (0x81)
#define SEN_CELL_MACRO                 (0x82)
#define SEN_CELL_CONTINUATION          (0x83)

typedef struct _sen_ql_co sen_ql_co;
typedef struct _sen_cell sen_cell;
typedef sen_cell *sen_ql_native_func(sen_ctx *, sen_cell *, sen_ql_co *);
typedef struct _sen_tmp_obj_db sen_tmp_obj_db;

struct _sen_cell {
  sen_obj_header header;
  union {
    struct {
      sen_id id;
      sen_ql_native_func *func;
    } o;
    struct {
      sen_obj *value;
      sen_ql_native_func *func;
    } p;
    struct {
      char *value;
      uint32_t size;
    } b;
    struct {
      sen_cell *car;
      sen_cell *cdr;
    } l;
    struct {
      int64_t i;
    } i;
    struct {
      double d;
    } d;
    struct {
      int8_t op;
      int8_t mode;
      int16_t weight;
      int32_t option;
    } op;
    sen_timeval tv;
  } u;
};

struct _sen_ql_co {
  uint16_t mode;
  uint16_t last;
  sen_ql_native_func *func;
  void *data;
};

struct _sen_tmp_obj_db {
  sen_obj_db *obj;
  sen_cell cell;
};

sen_cell *sen_get(const char *key);
sen_cell *sen_at(const char *key);
sen_rc sen_del(const char *key);

/**** sen_ctx_impl ****/

struct _sen_ctx_impl {
  sen_encoding encoding;

  /* memory pool portion */
  int32_t lifoseg;
  int32_t currseg;
  sen_io_mapinfo mi;
  sen_io_mapinfo *segs;

#ifdef USE_DYNAMIC_MALLOC_CHANGE
  /* memory allocation portion */
  sen_malloc_func malloc_func;
  sen_calloc_func calloc_func;
  sen_realloc_func realloc_func;
  sen_strdup_func strdup_func;
#endif

  /* ql portion */
  uint32_t ncells;
  uint32_t seqno;
  uint32_t lseqno;
  uint32_t nbinds;
  uint32_t nunbinds;
  uint8_t feed_mode;
  uint8_t batchmode;
  uint8_t gc_verbose;
  uint8_t op;
  int tok;
  char *cur;
  char *str_end;
  sen_cell **pht;       /* tail of placeholders */
  sen_cell arg;         /* wait_data container (for coroutine) */
  sen_db *db;
  uint32_t n_entries;
  sen_array *objects;       /* objects */
  sen_array *values;        /* temporary objects */
  sen_hash *symbols;        /* symbol table */
  sen_cell *phs;        /* list of placeholders */
  sen_ql_co co;             /* coroutine info */
  sen_cell *args;       /* register for arguments of function */
  sen_cell *envir;      /* stack register for current environment */
  sen_cell *code;       /* register for current code */
  sen_cell *dump;       /* stack register for next evaluation */
  sen_cell *value;      /* evaluated value */
  sen_cell *global_env; /* global variables */
  char *inbuf;
  sen_obj outbuf;
  sen_obj subbuf;
  unsigned int bufcur;
  void (*output)(sen_ctx *, int, void *);
  sen_com_sqtp *com;
  unsigned int com_status;
  unsigned int com_info;
  union {
    void *ptr;
    int fd;
    uint32_t u32;
    uint64_t u64;
  } data;
};

/**** query language ****/

#define SEN_QL_CO_BEGIN(c) if (c) { switch((c)->last) { case 0:
#define SEN_QL_CO_WAIT(c,d) \
  (c)->last = __LINE__; (c)->data = (d); return NULL;\
  case __LINE__: (d) = (c)->data;
#define SEN_QL_CO_END(c) (c)->last = 0; }}

#define SEN_QL_PLACEHOLDER_CHAR '?'

#define SEN_QL_SET_MODE(c,m) ((c)->impl->feed_mode = (m))
#define SEN_QL_GET_MODE(c) ((c)->impl->feed_mode)
#define SEN_QL_GET_STAT(c) ((c)->stat)

enum {
  sen_ql_atonce = 0,
  sen_ql_step
};

#define SEN_QL_TOPLEVEL  0x00
#define SEN_QL_QUITTING  0x0f
#define SEN_QL_EVAL      0x40
#define SEN_QL_NATIVE    0x41
#define SEN_QL_WAIT_EXPR 0xc0
#define SEN_QL_WAIT_ARG  0xc1
#define SEN_QL_WAIT_DATA 0xc2

#define SEN_QL_WAITINGP(c) ((c)->stat & 0x80)

extern sen_cell *sen_ql_nil;  /* special cell representing empty cell */
extern sen_cell *sen_ql_t;    /* special cell representing #t */
extern sen_cell *sen_ql_f;    /* special cell representing #f */

#define NIL sen_ql_nil
#define T sen_ql_t
#define F sen_ql_f

void sen_ql_init_globals(sen_ctx *c);

sen_cell *sen_ql_at(sen_ctx *c, const char *key);
sen_rc sen_ql_def_db_funcs(sen_ctx *c);

#define SEN_OBJ_INSPECT_ESC 1
#define SEN_OBJ_INSPECT_SYMBOL_AS_STR 2

void sen_obj_inspect(sen_ctx *c, sen_cell *obj, sen_obj *buf, int flags);
void sen_ql_def_native_func(sen_ctx *c, const char *name, sen_ql_native_func *func);
sen_cell *sen_ql_feed(sen_ctx *c, char *str, uint32_t str_size, int mode);
sen_cell *sen_ql_eval(sen_ctx *c, sen_cell *code, sen_cell *objs);
sen_rc sen_obj2int(sen_ctx *c, sen_cell *o);

const char *_sen_ql_obj_key(sen_ctx *c, sen_cell *obj, uint32_t *key_size);
sen_obj *sen_ql_obj_key(sen_ctx *ctx, sen_cell *obj, sen_obj *value);
sen_cell *sen_ql_mk_symbol(sen_ctx *ctx, const char *name, int name_size);
sen_cell *sen_ql_obj_new(sen_ctx *c, sen_id domain, sen_id self);
void sen_ql_obj_bind(sen_obj *obj, sen_cell *symbol);

sen_cell *sen_ql_mk_string(sen_ctx *c, const char *str, unsigned int len);
sen_cell *sen_ql_mk_symbol2(sen_ctx *ctx, const char *q, unsigned int len, int kwdp);

sen_cell *sen_cell_new(sen_ctx *c);
sen_cell *sen_cell_alloc(sen_ctx *c, uint32_t size);
void sen_cell_clear(sen_ctx *c, sen_cell *o);
sen_cell *sen_cell_cons(sen_ctx *ctx, sen_cell *a, sen_cell *b);
void sen_ql_init_const(void);

sen_rc sen_ql_obj_mark(sen_ctx *ctx, sen_cell *o);
sen_rc sen_ql_obj_unmark(sen_ctx *ctx, sen_cell *o);

typedef struct _patsnip_spec patsnip_spec;
void sen_obj_patsnip_spec_close(sen_ctx *ctx, patsnip_spec *ss);

#define SEN_OBJ2VALUE(o,v,s) ((v) = (o)->u.b.value, (s) = (o)->u.b.size)
#define SEN_VALUE2OBJ(o,v,s) ((o)->u.b.value = (v), (o)->u.b.size = (s))

#define VOIDP(c) ((c) == NIL || !(c) || (c)->header.type == SEN_VOID)
#define OBJECTP(c) ((c)->header.type == SEN_CELL_OBJECT)
#define RECORDSP(c) ((c)->header.type == SEN_TABLE_HASH_KEY)
#define SNIPP(c) ((c)->header.type == SEN_SNIP)
#define BULKP(c) ((c)->header.type == SEN_CELL_STR)
#define PAIRP(c) (((c)->header.type == SEN_CELL_LIST))
#define LISTP(c) (PAIRP(c) || (c) == NIL)
#define INTP(c) ((c)->header.type == SEN_CELL_INT)
#define TABLEP(c) (SEN_TABLE_HASH_KEY <= (c)->header.type && (c)->header.type <= SEN_SPACE)
#define RAW_TABLEP(c) ((c)->header.type == SEN_TYPE)
#define NATIVE_FUNCP(c) ((c)->header.impl_flags & SEN_CELL_NATIVE)
#define QUERYP(c) ((c)->header.type == SEN_QUERY)
#define RA_COLUMNP(c) ((c)->header.type == SEN_COLUMN_FIX_SIZE)
#define JA_COLUMNP(c) ((c)->header.type == SEN_COLUMN_VAR_SIZE)
#define IDX_COLUMNP(c) ((c)->header.type == SEN_COLUMN_INDEX)
#define COLUMNP(c) (RA_COLUMNP(c) || JA_COLUMNP(c) || IDX_COLUMNP(c))
#define OPP(c) ((c)->header.type == SEN_CELL_OP)
#define STRVALUE(c) ((c)->u.b.value)
#define STRSIZE(c) ((c)->u.b.size)
#define NUMBERP(c) ((c)->header.type == SEN_CELL_INT || (c)->header.type == SEN_CELL_FLOAT)
#define SYMBOLP(c) ((c)->header.impl_flags & SEN_CELL_SYMBOL)
#define CONTINUATIONP(p) ((p)->header.type == SEN_CELL_CONTINUATION)
#define CLOSUREP(p) ((p)->header.type == SEN_CELL_CLOSURE || (p)->header.type == SEN_CELL_MACRO)
#define PROCP(p) ((p)->header.type == SEN_CELL_PROC)
#define PROCEDUREP(p) (PROCP(p) || CLOSUREP(p) || CONTINUATIONP(p) || NATIVE_FUNCP(p))

#define SETINT(c,v) ((c)->header.type = SEN_CELL_INT, (c)->u.i.i = (v))
#define SETFLOAT(c,v) ((c)->header.type = SEN_CELL_FLOAT, (c)->u.d.d = (v))
#define SETBULK(c,v,s) ((c)->header.type = SEN_CELL_STR, (c)->u.b.value = (v), (c)->u.b.size = (s))
#define SETTIME(c,v) ((c)->header.type = SEN_CELL_TIME, memcpy(&(c)->u.tv, v, sizeof(sen_timeval)))

#define CAR(c)          ((c)->u.l.car)
#define CDR(c)          ((c)->u.l.cdr)
#define CAAR(c)         CAR(CAR(c))
#define CADR(c)         CAR(CDR(c))
#define CDAR(c)         CDR(CAR(c))
#define CDDR(c)         CDR(CDR(c))
#define CADDR(c)        CAR(CDDR(c))
#define CADAR(p)        CAR(CDAR(p))
#define CADAAR(p)       CAR(CDR(CAAR(p)))
#define CADDDR(p)       CAR(CDR(CDDR(p)))
#define CDDDDR(p)       CDR(CDR(CDDR(p)))

#define IVALUE(p)       ((p)->u.i.i)
#define FVALUE(p)       ((p)->u.d.d)

#define SEN_CELL_NEW(ctx,o) do {\
  if (!((o) = sen_cell_new(ctx))) { QLERR("obj_new failed"); }\
} while(0)

#define SEN_STR2OBJ(ctx,bulk,o) do {\
  SEN_BULK_PUTC(ctx, (bulk), '\0'); \
  SEN_CELL_NEW(ctx, (o));\
  (o)->header.impl_flags = SEN_OBJ_ALLOCATED;\
  (o)->header.type = SEN_CELL_STR;\
  (o)->u.b.value = (bulk)->u.b.head;\
  (o)->u.b.size = SEN_BULK_VSIZE(bulk) - 1;\
} while(0)

#define CONS(a,b) (sen_cell_cons(ctx, a, b))

#define INTERN(s) (sen_ql_mk_symbol(ctx, s, strlen(s)))

#define POP(x,c) (PAIRP(c) ? ((x) = CAR(c), (c) = CDR(c), (x)) : (x = NIL))

#define SKIPSPACE(c) do {\
  unsigned int len;\
  while ((c)->cur < (c)->str_end && sen_isspace((c)->cur, (c)->encoding)) {\
    if (!(len = sen_str_charlen_nonnull((c)->cur, (c)->str_end, (c)->encoding))) {\
      (c)->cur = (c)->str_end;\
      break;\
    }\
    (c)->cur += len;\
  }\
} while (0)

#ifdef __cplusplus
}
#endif

#endif /* SEN_QL_H */
