/* Copyright(C) 2004-2007 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_STORE_H
#define SEN_STORE_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifndef SEN_CTX_H
#include "ctx.h"
#endif /* SEN_CTX_H */

#ifndef SEN_HASH_H
#include "hash.h"
#endif /* SEN_HASH_H */

#ifndef SEN_IO_H
#include "io.h"
#endif /* SEN_IO_H */

#ifndef SEN_DB_H
#include "db.h"
#endif /* SEN_DB_H */

#ifdef __cplusplus
extern "C" {
#endif

#define SEN_ST_APPEND 1

/**** fixed sized elements ****/

typedef struct _sen_ra sen_ra;

struct _sen_ra {
  sen_obj_db obj;
  sen_io *io;
  int element_width;
  int element_mask;
  struct sen_ra_header *header;
};

struct sen_ra_header {
  char idstr[16];
  unsigned element_size;
  sen_id curr_max;
  uint32_t nrecords; /* nrecords is not maintained by default */
  uint32_t reserved[9];
};

sen_ra *sen_ra_create(sen_ctx *ctx, const char *path, unsigned int element_size);
sen_ra *sen_ra_open(sen_ctx *ctx, const char *path);
sen_rc sen_ra_info(sen_ctx *ctx, sen_ra *ra, unsigned int *element_size, sen_id *curr_max);
sen_rc sen_ra_close(sen_ctx *ctx, sen_ra *ra);
sen_rc sen_ra_remove(sen_ctx *ctx, const char *path);
void *sen_ra_get(sen_ctx *ctx, sen_ra *ra, sen_id id);
void *sen_ra_at(sen_ctx *ctx, sen_ra *ra, sen_id id);

/**** variable sized elements ****/

typedef struct _sen_ja sen_ja;

struct _sen_ja {
  sen_obj_db obj;
  sen_io *io;
  struct sen_ja_header *header;
  uint32_t *dsegs;
  uint32_t *esegs;
};

sen_ja *sen_ja_create(sen_ctx *ctx, const char *path,
                      unsigned int max_element_size, uint32_t flags);
sen_ja *sen_ja_open(sen_ctx *ctx, const char *path);
sen_rc sen_ja_info(sen_ctx *ctx, sen_ja *ja, unsigned int *max_element_size);
sen_rc sen_ja_close(sen_ctx *ctx, sen_ja *ja);
sen_rc sen_ja_remove(sen_ctx *ctx, const char *path);
sen_rc sen_ja_put(sen_ctx *ctx, sen_ja *ja, sen_id id,
                  void *value, int value_len, int flags);
int sen_ja_at(sen_ctx *ctx, sen_ja *ja, sen_id id, void *valbuf, int buf_size);

void *sen_ja_ref(sen_ctx *ctx, sen_ja *ja, sen_id id, uint32_t *value_len);
sen_rc sen_ja_unref(sen_ctx *ctx, sen_ja *ja, sen_id id, void *value, uint32_t value_len);
int sen_ja_defrag(sen_ctx *ctx, sen_ja *ja, int threshold);

sen_rc sen_ja_putv(sen_ctx *ctx, sen_ja *ja, sen_id id, sen_obj *vector, int flags);

/**** vgram ****/

typedef struct _sen_vgram_vnode
{
  struct _sen_vgram_vnode *car;
  struct _sen_vgram_vnode *cdr;
  sen_id tid;
  sen_id vid;
  int freq;
  int len;
} sen_vgram_vnode;

struct _sen_vgram {
  sen_sym *vgram;
};

struct _sen_vgram_buf {
  size_t len;
  sen_id *tvs;
  sen_id *tvp;
  sen_id *tve;
  sen_vgram_vnode *vps;
  sen_vgram_vnode *vpp;
  sen_vgram_vnode *vpe;
};

sen_vgram *sen_vgram_create(const char *path);
sen_vgram *sen_vgram_open(const char *path);
sen_rc sen_vgram_close(sen_vgram *vgram);
sen_rc sen_vgram_update(sen_vgram *vgram, sen_id rid, sen_vgram_buf *b, sen_hash *terms);

sen_vgram_buf *sen_vgram_buf_open(size_t len);
sen_rc sen_vgram_buf_add(sen_vgram_buf *b, sen_id tid);
sen_rc sen_vgram_buf_close(sen_vgram_buf *b);

#ifdef __cplusplus
}
#endif

#endif /* SEN_STORE_H */
