/* Copyright(C) 2004-2008 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef SEN_STR_H
#define SEN_STR_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifndef SEN_NFKC_H
#include "nfkc.h"
#endif /* SEN_NFKC_H */

#ifdef	__cplusplus
extern "C" {
#endif

typedef struct {
  const char *orig;
  size_t orig_blen;
  char *norm;
  size_t norm_blen;
  uint_least8_t *ctypes;
  int16_t *checks;
  size_t length;
  int flags;
  sen_ctx *ctx;
  sen_encoding encoding;
} sen_nstr;

typedef enum {
  getopt_op_none = 0,
  getopt_op_on,
  getopt_op_off,
  getopt_op_update
} sen_str_getopt_op;

typedef struct {
  const char opt; /* ends opt == 0 && longopt == NULL */
  const char *longopt;
  char **arg; /* if NULL, no arg are required */
  int flag;
  sen_str_getopt_op op;
} sen_str_getopt_opt;

sen_nstr *sen_nstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags);
sen_nstr *sen_fakenstr_open(const char *str, size_t str_len, sen_encoding encoding, int flags);
sen_rc sen_nstr_close(sen_nstr *nstr);

size_t sen_str_charlen_nonnull(const char *str, const char *end, sen_encoding encoding);
size_t sen_str_len(const char *str, sen_encoding encoding, const char **last);

#define SEN_NSTR_BLANK 0x80
#define SEN_NSTR_ISBLANK(c) (c & 0x80)
#define SEN_NSTR_CTYPE(c) (c & 0x7f)

int sen_isspace(const char *s, sen_encoding encoding);
int sen_atoi(const char *nptr, const char *end, const char **rest);
unsigned int sen_atoui(const char *nptr, const char *end, const char **rest);
unsigned int sen_htoui(const char *nptr, const char *end, const char **rest);
int64_t sen_atoll(const char *nptr, const char *end, const char **rest);
sen_rc sen_itoa(int i, char *p, char *end, char **rest);
sen_rc sen_lltoa(int64_t i, char *p, char *end, char **rest);
const char *sen_enctostr(sen_encoding enc);
sen_encoding sen_strtoenc(const char *str);

void sen_itoh(unsigned int i, char *p, unsigned int len);
int sen_str_tok(char *str, size_t str_len, char delim, char **tokbuf, int buf_size, char **rest);
int sen_str_getopt(int argc, char * const argv[], const sen_str_getopt_opt *opts, int *flags);

extern int sen_str_margin_size;

char *sen_itob(sen_id id, char *p);
sen_id sen_btoi(char *b);

sen_rc sen_substring(char **str, char **str_end, int start, int end, sen_encoding encoding);

void sen_logger_fin(void);

/* bulk */

#define SEN_BULK_VALUE(bulk) \
 (((bulk)->header.type == SEN_BULK) ? SEN_BULK_HEAD(bulk) : NULL)

#define SEN_BULK_LEN(bulk) \
 (((bulk)->header.type == SEN_BULK) ? SEN_BULK_VSIZE(bulk) : 0)

#define SEN_BULK_SET(ctx,bulk,str,len) {\
  if ((bulk)->header.type == SEN_VOID) {\
    SEN_OBJ_INIT((bulk), SEN_BULK, 0);\
  }\
  if ((bulk)->header.type == SEN_BULK) {\
    if ((bulk)->header.flags & SEN_OBJ_DO_SHALLOW_COPY) {\
      (bulk)->u.b.head = (char *)(str);\
      (bulk)->u.b.curr = (char *)(str) + len;\
    } else {\
      sen_bulk_write((ctx), (bulk), (const char *)(str), (unsigned int)(len));\
    }\
  } else {\
    (ctx)->rc = sen_invalid_argument;\
  }\
}

#ifdef __cplusplus
}
#endif

#endif /* SEN_STR_H */
