/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "senna_in.h"
#include <string.h>
#include <ctype.h>
#include "token.h"

/* ngram */

inline static sen_token *
sen_ngram_init(sen_token *token)
{
  return token;
}

inline static sen_id
sen_ngram_next(sen_token *token)
{
  sen_id tid;
  sen_obj *table = token->table;
  sen_ctx *ctx = token->ctx;
  uint_least8_t *cp = NULL;
  int32_t len = 0, pos;
  const unsigned char *p, *q, *r;
  if (token->status == sen_token_done) { return SEN_ID_NIL; }
  token->force_prefix = 0;
  for (p = token->next, pos = token->pos + token->skip; *p; p = r, pos++) {
    if (token->nstr->ctypes) { cp = token->nstr->ctypes + pos; }
    if (token->uni_alpha && SEN_NSTR_CTYPE(*cp) == sen_str_alpha) {
      for (len = 1, r = p;;len++) {
        size_t cl;
        if (!(cl = sen_str_charlen((char *)r, token->encoding))) { break; }
        r += cl;
        if (SEN_NSTR_ISBLANK(*cp)) { break; }
        if (SEN_NSTR_CTYPE(*++cp) != sen_str_alpha) { break; }
      }
      {
        size_t blen = r - p;
        if (!blen) {
          token->status = sen_token_done;
          return SEN_ID_NIL;
        }
        token->curr = p;
        token->curr_size = blen;
        tid = sen_table_lookup(ctx, table, p, blen, &token->flags);
        token->skip = len;
      }
    } else if (token->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) {
      for (len = 1, r = p;;len++) {
        size_t cl;
        if (!(cl = sen_str_charlen((char *)r, token->encoding))) { break; }
        r += cl;
        if (SEN_NSTR_ISBLANK(*cp)) { break; }
        if (SEN_NSTR_CTYPE(*++cp) != sen_str_digit) { break; }
      }
      {
        size_t blen = r - p;
        if (!blen) {
          token->status = sen_token_done;
          return SEN_ID_NIL;
        }
        token->curr = p;
        token->curr_size = (uint32_t)blen;
        tid = sen_table_lookup(ctx, table, p, blen, &token->flags);
        token->skip = len;
      }
    } else if (token->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol) {
      for (len = 1, r = p;;len++) {
        size_t cl;
        if (!(cl = sen_str_charlen((char *)r, token->encoding))) { break; }
        r += cl;
        if (SEN_NSTR_ISBLANK(*cp)) { break; }
        if (SEN_NSTR_CTYPE(*++cp) != sen_str_symbol) { break; }
      }
      {
        size_t blen = r - p;
        if (!blen) {
          token->status = sen_token_done;
          return SEN_ID_NIL;
        }
        token->curr = p;
        token->curr_size = (uint32_t)blen;
        tid = sen_table_lookup(ctx, table, p, blen, &token->flags);
        token->skip = len;
      }
    } else {
      size_t cl;
#ifdef PRE_DEFINED_UNSPLIT_WORDS
      {
        const unsigned char *key = NULL;
        // todo : sen_pat_lcp_search
        if ((tid = sen_sym_common_prefix_search(sym, p))) {
          if (!(key = _sen_sym_key(sym, tid))) {
            token->status = sen_token_not_found;
            return SEN_ID_NIL;
          }
          len = sen_str_len(key, token->encoding, NULL);
        }
        r = p + sen_str_charlen(p, token->encoding);
        if (tid && (len > 1 || r == p)) {
          if (r != p && pos + len - 1 <= token->tail) { continue; }
          p += strlen(key);
          if (!*p && !(token->flags & SEN_TABLE_ADD)) { token->status = sen_token_done; }
        }
      }
#endif /* PRE_DEFINED_UNSPLIT_WORDS */
      if (!(cl = sen_str_charlen((char *)p, token->encoding))) {
        token->status = sen_token_done;
        return SEN_ID_NIL;
      }
      r = p + cl;
      {
        int blankp = 0;
        for (len = 1, q = r; len < token->ngram_unit; len++) {
          if (cp) {
            if (SEN_NSTR_ISBLANK(*cp)) { blankp++; break; }
            cp++;
          }
          if (!(cl = sen_str_charlen((char *)q, token->encoding)) ||
              (token->uni_alpha && SEN_NSTR_CTYPE(*cp) == sen_str_alpha) ||
              (token->uni_digit && SEN_NSTR_CTYPE(*cp) == sen_str_digit) ||
              (token->uni_symbol && SEN_NSTR_CTYPE(*cp) == sen_str_symbol)) {
            break;
          }
          q += cl;
        }
        if (blankp && !(token->flags & SEN_TABLE_ADD)) { continue; }
      }
      if ((!cl || !*q) && !(token->flags & SEN_TABLE_ADD)) { token->status = sen_token_done; }
      if (len < token->ngram_unit) { token->force_prefix = 1; }
      {
        size_t blen = q - p;
        if (!blen) {
          token->status = sen_token_done;
          return SEN_ID_NIL;
        }
        token->curr = p;
        token->curr_size = (uint32_t)blen;
        tid = sen_table_lookup(ctx, table, p, blen, &token->flags);
        token->skip = 1;
      }
    }
    token->pos = pos;
    token->len = len;
    token->tail = pos + len - 1;
    token->next = r;
    // printf("tid=%d pos=%d tail=%d (%s) %s\n", tid, token->pos, token->tail, _sen_sym_key(sym, tid), r);
    // printf("tid=%d pos=%d tail=%d (%s)\n", tid, token->pos, token->tail, _sen_sym_key(sym, tid));
    if (!tid) {
      token->status = sen_token_not_found;
    } else {
      if (!*r) { token->status = sen_token_done; }
    }
    return tid;
  }
  token->status = sen_token_done;
  return SEN_ID_NIL;
}

/* mecab */

#ifndef NO_MECAB

static mecab_t *sole_mecab;
static sen_mutex sole_mecab_lock;

static char *sen_token_default_mecab_argv[] = {"", "-Owakati"};

static int sen_token_mecab_argc = 2;
static char **sen_token_mecab_argv = sen_token_default_mecab_argv;

#define SOLE_MECAB_CONFIRM do {\
  if (!sole_mecab) {\
    MUTEX_LOCK(sole_mecab_lock);\
    if (!sole_mecab) { sole_mecab = mecab_new(sen_token_mecab_argc, sen_token_mecab_argv); }\
    MUTEX_UNLOCK(sole_mecab_lock);\
  }\
} while(0)

inline static sen_token *
sen_mecab_init(sen_token *token)
{
  sen_ctx *ctx = token->ctx;
  unsigned int bufsize, maxtrial = 10, len;
  sen_nstr *nstr = token->nstr;
  char *buf, *s, *p;
  char mecab_err[256];
  // sen_log("(%s)", str);
  SOLE_MECAB_CONFIRM;
  if (!sole_mecab) {
    SEN_LOG(sen_log_alert, "mecab_new failed on sen_mecab_init");
    return NULL;
  }
  token->mecab = sole_mecab;
  // if (!(token->mecab = mecab_new3())) {
  len = nstr->norm_blen;
  mecab_err[sizeof(mecab_err) - 1] = '\0';
  for (bufsize = len * 2 + 1; maxtrial; bufsize *= 2, maxtrial--) {
    if(!(buf = SEN_MALLOC(bufsize + 1))) {
      SEN_LOG(sen_log_alert, "buffer allocation on sen_mecab_init failed !");
      SEN_FREE(token);
      return NULL;
    }
    MUTEX_LOCK(sole_mecab_lock);
    s = mecab_sparse_tostr3(token->mecab, (char *)nstr->norm, len, buf, bufsize);
    if (!s) {
      strncpy(mecab_err, mecab_strerror(token->mecab), sizeof(mecab_err) - 1);
    }
    MUTEX_UNLOCK(sole_mecab_lock);
    if (s) { break; }
    SEN_FREE(buf);
    if (strstr(mecab_err, "output buffer overflow") == NULL) {
      break;
    }
  }
  if (!s) {
    SEN_LOG(sen_log_alert, "mecab_sparse_tostr failed len=%d bufsize=%d err=%s", len, bufsize, mecab_err);
    sen_token_close(token);
    return NULL;
  }
  // certain version of mecab returns trailing lf or spaces.
  for (p = buf + strlen(buf) - 1; buf <= p && (*p == '\n' || isspace(*(unsigned char *)p)); p--) { *p = '\0'; }
  //sen_log("sparsed='%s'", s);
  token->orig = (unsigned char *)nstr->norm;
  token->buf = (unsigned char *)buf;
  token->next = (unsigned char *)buf;
  token->force_prefix = 0;
  return token;
}

inline static sen_id
sen_mecab_next(sen_token *token)
{
  sen_id tid;
  sen_obj *table = token->table;
  sen_ctx *ctx = token->ctx;
  int32_t len, offset = token->offset + token->len;
  const unsigned char *p;
  if (token->status == sen_token_done) { return SEN_ID_NIL; }
  for (p = token->next, len = 0;;) {
    size_t cl;
    if (!(cl = sen_str_charlen((char *)p, token->encoding)) ||
        sen_isspace((const char *)p, token->encoding)) {
      break;
    }
    p += cl;
    len++;
  }
  if (!len) {
    token->status = sen_token_done;
    return SEN_ID_NIL;
  }
  token->curr = token->next;
  token->curr_size = (uint32_t)(p - token->next);
  tid = sen_table_lookup(ctx, table, token->curr, token->curr_size, &token->flags);
  {
    int cl;
    while ((cl = sen_isspace((const char *)p, token->encoding))) { p += cl; }
    token->next = p;
    token->offset = offset;
    token->len = len;
  }
  if (tid == SEN_ID_NIL) {
    token->status = sen_token_not_found;
  } else {
    if (!*p) { token->status = sen_token_done; }
  }
  token->pos++;
  return tid;
}

sen_rc
sen_token_set_mecab_args(int argc, char **argv)
{
  sen_token_mecab_argc = argc;
  sen_token_mecab_argv = argv;
  if (sole_mecab) {
    SEN_LOG(sen_log_alert, "mecab already initialized");
    return sen_invalid_argument;
  }
  SOLE_MECAB_CONFIRM;
  return sen_success;
}

#endif /* NO_MECAB */

/* delimited */

inline static sen_token *
sen_delimited_init(sen_token *token)
{
  int cl;
  const char *p = token->nstr->norm;
  const char *pe = token->nstr->norm + token->nstr->norm_blen;
  token->orig = (unsigned char *)p;
  while ((cl = sen_isspace((const char *)p, token->encoding))) {
    p += cl;
    if (pe <= p) {
      token->status = sen_token_done;
      break;
    }
  }
  token->next = (unsigned char *)p;
  return token;
}

inline static sen_id
sen_delimited_next(sen_token *token)
{
  sen_id tid;
  sen_obj *table = token->table;
  sen_ctx *ctx = token->ctx;
  int32_t len, offset = token->offset + token->len;
  const unsigned char *p;
  if (token->status == sen_token_done) { return SEN_ID_NIL; }
  for (p = token->next, len = 0;;) {
    size_t cl;
    if (!(cl = sen_str_charlen((char *)p, token->encoding)) ||
        sen_isspace((const char *)p, token->encoding)) {
      break;
    }
    p += cl;
    len++;
  }
  if (!len) {
    token->status = sen_token_done;
    return SEN_ID_NIL;
  }
  token->curr = token->next;
  token->curr_size = (uint32_t)(p - token->next);
  tid = sen_table_lookup(ctx, table, token->curr, token->curr_size, &token->flags);
  {
    int cl;
    while ((cl = sen_isspace((const char *)p, token->encoding))) { p += cl; }
    token->next = p;
    token->offset = offset;
    token->len = len;
  }
  if (tid == SEN_ID_NIL) {
    token->status = sen_token_not_found;
  } else {
    if (!*p) { token->status = sen_token_done; }
  }
  token->pos++;
  return tid;
}

/* external */

sen_rc
sen_token_init(void)
{
#ifndef NO_MECAB
  // char *arg[] = {"", "-Owakati"};
  // return mecab_load_dictionary(2, arg) ? sen_success : sen_external_error;
  sole_mecab = NULL;
  MUTEX_INIT(sole_mecab_lock);
#endif /* NO_MECAB */
  return sen_success;
}

sen_rc
sen_token_fin(void)
{
#ifndef NO_MECAB
  if (sole_mecab) {
    mecab_destroy(sole_mecab);
    sole_mecab = NULL;
  }
  MUTEX_DESTROY(sole_mecab_lock);
#endif /* NO_MECAB */
  return sen_success;
}

sen_token *
sen_token_open(sen_ctx *ctx, sen_obj *table, const char *str, size_t str_len,
               sen_search_flags flags)
{
  sen_nstr *nstr;
  int nflag, type;
  sen_token *token;
  sen_encoding encoding;
  sen_obj_flags table_flags;
  uint8_t ngram_unit;
  if (sen_table_get_info(ctx, table, &table_flags, &encoding, &ngram_unit)) { return NULL; }
  type = table_flags & SEN_OBJ_TOKEN_MASK;
  nflag = (type == SEN_OBJ_TOKEN_NGRAM ? SEN_STR_REMOVEBLANK|SEN_STR_WITH_CTYPES : 0);
  if (table_flags & SEN_OBJ_KEY_NORMALIZE) {
    if (!(nstr = sen_nstr_open(str, str_len, encoding, nflag))) {
      SEN_LOG(sen_log_alert, "sen_nstr_open failed at sen_token_open");
      return NULL;
    }
  } else {
    if (!(nstr = sen_fakenstr_open(str, str_len, encoding, nflag))) {
      SEN_LOG(sen_log_alert, "sen_fakenstr_open failed at sen_token_open");
      return NULL;
    }
  }
  if (!(token = SEN_MALLOC(sizeof(sen_token)))) { return NULL; }
  token->ctx = ctx;
  token->table = table;
  token->flags = flags;
  token->nstr = nstr;
  token->table_flags = table_flags;
  token->encoding = encoding;

#ifndef NO_MECAB
  token->mecab = NULL;
#endif /* NO_MECAB */
  token->buf = NULL;
  token->curr = NULL;
  token->curr_size = 0;
  token->pos = -1;
  token->skip = 1;
  token->tail = 0;
  token->status = sen_token_doing;

  token->orig = (unsigned char *)nstr->norm;
  token->orig_blen = nstr->norm_blen;
  token->next = (unsigned char *)nstr->norm;
  token->uni_alpha = (nstr->ctypes && !(table_flags & SEN_OBJ_KEY_SPLIT_ALPHA));
  token->uni_digit = (nstr->ctypes && !(table_flags & SEN_OBJ_KEY_SPLIT_DIGIT));
  token->uni_symbol = (nstr->ctypes && !(table_flags & SEN_OBJ_KEY_SPLIT_SYMBOL));
  token->force_prefix = 0;
  token->ngram_unit = ngram_unit;
  token->offset = 0;
  token->len = 0;

  switch (type) {
  case SEN_OBJ_TOKEN_MECAB :
#ifdef NO_MECAB
    SEN_FREE(token);
    return NULL;
#else /* NO_MECAB */
    return sen_mecab_init(token);
#endif /* NO_MECAB */
  case SEN_OBJ_TOKEN_NGRAM :
    return sen_ngram_init(token);
  case SEN_OBJ_TOKEN_DELIMITED :
    return sen_delimited_init(token);
  default :
    SEN_FREE(token);
    return NULL;
  }
}

sen_rc
sen_token_next(sen_token *token)
{
  /* if (!token) { return sen_invalid_argument; } */
  switch ((token->table_flags & SEN_OBJ_TOKEN_MASK)) {
  case SEN_OBJ_TOKEN_MECAB :
#ifdef NO_MECAB
    return sen_invalid_argument;
#else /* NO_MECAB */
    return sen_mecab_next(token);
#endif /* NO_MECAB */
  case SEN_OBJ_TOKEN_NGRAM :
    return sen_ngram_next(token);
  case SEN_OBJ_TOKEN_DELIMITED :
    return sen_delimited_next(token);
  default :
    return sen_invalid_argument;
  }
}

sen_rc
sen_token_close(sen_token *token)
{
  if (token) {
    sen_ctx *ctx = token->ctx;
    if (token->nstr) { sen_nstr_close(token->nstr); }
    // if (token->mecab) { mecab_destroy(token->mecab); }
    if (token->buf) { SEN_FREE(token->buf); }
    SEN_FREE(token);
    return sen_success;
  } else {
    return sen_invalid_argument;
  }
}

sen_rc
sen_token_validate(sen_ctx *ctx, sen_obj *table)
{
  sen_rc rc;
  sen_encoding encoding;
  sen_obj_flags table_flags;
  uint8_t ngram_unit;
  if ((rc = sen_table_get_info(ctx, table, &table_flags, &encoding, &ngram_unit))) {
    return rc;
  }
#ifndef NO_MECAB
#ifdef USE_MECAB_DICINFO
  if ((table_flags & SEN_OBJ_TOKEN_MASK) == SEN_OBJ_TOKEN_MECAB) {
    sen_encoding enc;
    const mecab_dictionary_info_t *di;

    SOLE_MECAB_CONFIRM;
    if (!sole_mecab) {
      SEN_LOG(sen_log_alert, "mecab_new failed on sen_token_validate");
      return sen_external_error;
    }
    di = mecab_dictionary_info(sole_mecab);
    if (!di || !di->charset) {
      SEN_LOG(sen_log_alert, "mecab_dictionary_info failed on sen_token_validate");
      return sen_external_error;
    }
    switch (di->charset[0]) {
      case 'u':
        enc = sen_enc_utf8;
        break;
      case 'e':
        enc = sen_enc_euc_jp;
        break;
      case 'c': /* cp932 */
      case 's':
        enc = sen_enc_sjis;
        break;
      default:
        SEN_LOG(sen_log_alert, "unknown encoding %s on sen_token_validate", di->charset);
        return sen_external_error;
    }
    if (enc != encoding) {
      SEN_LOG(sen_log_alert,
              "dictionary encoding %s is differ from table encoding %s",
              di->charset, sen_enctostr(encoding));
      return sen_abnormal_error;
    }
  }
#endif /* USE_MECAB_DICINFO */
#endif /* NO_MECAB */
  return sen_success;
}
