/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef NO_MECAB
#include <mecab.h>
#endif /* NO_MECAB */

#ifndef SEN_TOKEN_H
#define SEN_TOKEN_H

#ifndef SENNA_H
#include "senna_in.h"
#endif /* SENNA_H */

#ifndef SEN_CTX_H
#include "ctx.h"
#endif /* SEN_CTX_H */

#ifndef SEN_DB_H
#include "db.h"
#endif /* SEN_DB_H */

#ifndef SEN_STR_H
#include "str.h"
#endif /* SEN_STR_H */

#ifdef	__cplusplus
extern "C" {
#endif

#define SEN_TOKEN_ADD 1
#define SEN_TOKEN_UPD 2

typedef struct {
  sen_ctx *ctx;
  sen_obj *table;
  unsigned char *buf;
  const unsigned char *orig;
  const unsigned char *next;
  const unsigned char *curr;
  uint32_t orig_blen;
  uint32_t curr_size;
  sen_nstr *nstr;
#ifndef NO_MECAB
  mecab_t *mecab;
#endif /* NO_MECAB */
  int32_t pos;
  int32_t len;
  uint32_t skip;
  uint32_t tail;
  uint32_t offset;
  sen_search_flags flags;
  uint8_t status;
  uint8_t uni_alpha;
  uint8_t uni_digit;
  uint8_t uni_symbol;
  uint8_t force_prefix;
  uint8_t ngram_unit;
  sen_obj_flags table_flags;
  sen_encoding encoding;
} sen_token;

enum {
  sen_token_doing = 0,
  sen_token_done,
  sen_token_not_found
};

sen_rc sen_token_init(void);
sen_token *sen_token_open(sen_ctx *ctx, sen_obj *table, const char *str,
                          size_t str_len, sen_search_flags flags);

sen_rc sen_token_next(sen_token *ng);
sen_rc sen_token_close(sen_token *ng);
sen_rc sen_token_fin(void);
sen_rc sen_token_validate(sen_ctx *ctx, sen_obj *table);

#ifdef __cplusplus
}
#endif

#endif /* SEN_TOKEN_H */
